"""
kombu JSON-RPC server integration.
"""

import logging
from typing import Any, Dict, List, Optional, Type

import kombu.mixins
from kombu.transport import virtual

import pjrpc.server

logger = logging.getLogger(__package__)

KombuDispatcher = pjrpc.server.Dispatcher[kombu.Message]


class Executor(kombu.mixins.ConsumerProducerMixin):
    """
    `kombu <http://kombu.readthedocs.org/>`_ based JSON-RPC server.

    :param broker_url: broker connection url
    :param queue_name: requests queue name
    :param conn_args: additional connection arguments
    :param queue_args: queue arguments
    :param publish_args: message publish additional arguments
    :param prefetch_count: worker prefetch count
    :param kwargs: dispatcher additional arguments
    """

    def __init__(
        self,
        broker_url: str,
        queue_name: str,
        conn_args: Optional[Dict[str, Any]] = None,
        queue_args: Optional[Dict[str, Any]] = None,
        publish_args: Optional[Dict[str, Any]] = None,
        prefetch_count: int = 0,
        **kwargs: Any,
    ):
        self.connection = kombu.Connection(broker_url, **(conn_args or {}))

        self._rpc_queue = kombu.Queue(queue_name, **(queue_args or {}))
        self._prefetch_count = prefetch_count
        self._publish_args = publish_args

        self._dispatcher = KombuDispatcher(**kwargs)

    @property
    def dispatcher(self) -> KombuDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    def get_consumers(self, Consumer: Type[kombu.Consumer], channel: virtual.AbstractChannel) -> List[kombu.Consumer]:
        return [
            Consumer(
                channel=channel,
                queues=[self._rpc_queue],
                on_message=self._rpc_handle,
                accept={pjrpc.common.DEFAULT_CONTENT_TYPE},
                prefetch_count=self._prefetch_count,
            ),
        ]

    def _rpc_handle(self, message: kombu.Message) -> None:
        """
        Handles JSON-RPC request.

        :param message: kombu message :py:class:`kombu.message.Message`
        """

        try:
            reply_to = message.properties.get('reply_to')
            response = self._dispatcher.dispatch(message.body, context=message)

            if response is not None:
                response_text, error_codes = response
                if reply_to is None:
                    logger.warning("property 'reply_to' is missing")
                else:
                    self.producer.publish(
                        response_text,
                        routing_key=reply_to,
                        correlation_id=message.properties.get('correlation_id'),
                        content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                        content_encoding='utf8',
                        **(self._publish_args or {}),
                    )

            message.ack()

        except Exception as e:
            logger.exception("jsonrpc request handling error: %s", e)



"""
JSON-RPC server package.
"""

from . import typedefs
from .dispatcher import AsyncDispatcher, BaseDispatcher, Dispatcher, JSONEncoder, Method, MethodRegistry, ViewMixin

__all__ = [
    'AsyncDispatcher',
    'BaseDispatcher',
    'Dispatcher',
    'JSONEncoder',
    'Method',
    'MethodRegistry',
    'ViewMixin',
    'typedefs',
]



"""
OpenAPI Specification generator. See https://swagger.io/specification/.
"""

try:
    import dataclasses as dc
except ImportError:
    raise AssertionError("python 3.7 or later is required")

import copy
import enum
import functools as ft
import pathlib
import re
from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Type, TypedDict, Union

from pjrpc.common import UNSET, MaybeSet, UnsetType, exceptions
from pjrpc.common.typedefs import Func
from pjrpc.server import Method, utils
from pjrpc.server.specs.schemas import build_request_schema, build_response_schema

from . import BaseUI, Specification, extractors

HTTP_DEFAULT_STATUS = 200
JSONRPC_MEDIATYPE = 'application/json'


def drop_unset(obj: Any) -> Any:
    if isinstance(obj, dict):
        return dict((drop_unset(k), drop_unset(v)) for k, v in obj.items() if k is not UNSET and v is not UNSET)
    if isinstance(obj, (tuple, list, set)):
        return list(drop_unset(v) for v in obj if v is not UNSET)

    return obj


JsonSchema = Dict[str, Any]


@dc.dataclass
class Reference:
    """
    A simple object to allow referencing other components in the OpenAPI document, internally and externally.

    :param ref: the reference identifier.
    :param summary: a short summary which by default SHOULD override that of the referenced component.
    :param description: a description which by default SHOULD override that of the referenced component
    """

    ref: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET

    def __post_init__(self) -> None:
        self.__dict__['$ref'] = self.__dict__['ref']
        self.__dataclass_fields__['ref'].name = '$ref'  # noqa


@dc.dataclass
class Contact:
    """
    Contact information for the exposed API.

    :param name: the identifying name of the contact person/organization
    :param url: the URL pointing to the contact information
    :param email: the email address of the contact person/organization
    """

    name: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET
    email: MaybeSet[str] = UNSET


@dc.dataclass
class License:
    """
    License information for the exposed API.

    :param name: the license name used for the API
    :param identifier: an SPDX license expression for the API
    :param url: a URL to the license used for the API
    """

    name: str
    identifier: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET


@dc.dataclass
class Info:
    """
    Metadata about the API.

    :param title: the title of the application
    :param version: the version of the OpenAPI document
    :param summary: a short summary of the API.
    :param description: a description of the application
    :param contact: the contact information for the exposed API
    :param license: the license information for the exposed API
    :param termsOfService: a URL to the Terms of Service for the API
    """

    title: str
    version: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    contact: MaybeSet[Contact] = UNSET
    license: MaybeSet[License] = UNSET
    termsOfService: MaybeSet[str] = UNSET


@dc.dataclass
class ServerVariable:
    """
    An object representing a Server Variable for server URL template substitution.

    :param default: the default value to use for substitution, which SHALL be sent if an alternate value is not supplied
    :param enum: an enumeration of string values to be used if the substitution options are from a limited set
    :param description: an optional description for the server variable
    """

    default: str
    enum: MaybeSet[List[str]] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Server:
    """
    Connectivity information of a target server.

    :param url: a URL to the target host
    :param description: an optional string describing the host designated by the URL
    :param variables: a map between a variable name and its value.
                      The value is used for substitution in the server's URL template.
    """

    url: str
    description: MaybeSet[str] = UNSET
    variables: MaybeSet[Dict[str, ServerVariable]] = UNSET


@dc.dataclass
class Link:
    """
    The Link object represents a possible design-time link for a response.

    :param operationRef: a relative or absolute URI reference to an OAS operation
    :param operationId: the name of an existing, resolvable OAS operation, as defined with a unique operationId
    :param parameters: a map representing parameters to pass to an operation as specified with operationId
                       or identified via operationRef
    :param requestBody: a literal value or {expression} to use as a request body when calling the target operation
    :param description: a description of the link
    :param server: a server object to be used by the target operation.
    """

    operationRef: MaybeSet[str] = UNSET
    operationId: MaybeSet[str] = UNSET
    parameters: MaybeSet[Dict[str, Any]] = UNSET
    requestBody: MaybeSet[Any] = UNSET
    description: MaybeSet[str] = UNSET
    server: MaybeSet[Server] = UNSET


@dc.dataclass
class ExternalDocumentation:
    """
    Allows referencing an external resource for extended documentation.

    :param url: a short description of the target documentation.
    :param description: the URL for the target documentation
    """

    url: str
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Tag:
    """
    A list of tags for API documentation control.
    Tags can be used for logical grouping of methods by resources or any other qualifier.

    :param name: the name of the tag
    :param externalDocs: additional external documentation for this tag
    :param description: a short description for the tag
    """

    name: str
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET


class SecuritySchemeType(str, enum.Enum):
    """
    The type of the security scheme.
    """

    APIKEY = 'apiKey'
    HTTP = 'http'
    OAUTH2 = 'oauth2'
    OPENID_CONNECT = 'openIdConnect'


class ApiKeyLocation(str, enum.Enum):
    """
    The location of the API key.
    """

    QUERY = 'query'
    HEADER = 'header'
    COOKIE = 'cookie'


@dc.dataclass
class OAuthFlow:
    """
    Configuration details for a supported OAuth Flow.

    :param authorizationUrl: the authorization URL to be used for this flow
    :param tokenUrl: the token URL to be used for this flow
    :param refreshUrl: the URL to be used for obtaining refresh tokens
    :param scopes: the available scopes for the OAuth2 security scheme
    """

    authorizationUrl: str
    tokenUrl: str
    scopes: Dict[str, str]
    refreshUrl: MaybeSet[str] = UNSET


@dc.dataclass
class OAuthFlows:
    """
    Configuration of the supported OAuth Flows.

    :param implicit: configuration for the OAuth Implicit flow
    :param password: configuration for the OAuth Resource Owner Password flow
    :param clientCredentials: configuration for the OAuth Client Credentials flow
    :param authorizationCode: configuration for the OAuth Authorization Code flow
    """

    implicit: MaybeSet[OAuthFlow] = UNSET
    password: MaybeSet[OAuthFlow] = UNSET
    clientCredentials: MaybeSet[OAuthFlow] = UNSET
    authorizationCode: MaybeSet[OAuthFlow] = UNSET


@dc.dataclass
class SecurityScheme:
    """
    Defines a security scheme that can be used by the operations.

    :param type: the type of the security scheme
    :param name: the name of the header, query or cookie parameter to be used
    :param location: the location of the API key
    :param scheme: the name of the HTTP Authorization scheme to be used in the Authorization header
    :param bearerFormat: a hint to the client to identify how the bearer token is formatted
    :param flows: an object containing configuration information for the flow types supported
    :param openIdConnectUrl:
    :param description: a short description for security scheme
    """

    type: SecuritySchemeType
    scheme: MaybeSet[str] = UNSET
    name: MaybeSet[str] = UNSET
    location: MaybeSet[ApiKeyLocation] = UNSET  # `in` field
    bearerFormat: MaybeSet[str] = UNSET
    flows: MaybeSet[OAuthFlows] = UNSET
    openIdConnectUrl: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET

    def __post_init__(self) -> None:
        # `in` field name is not allowed in python
        self.__dict__['in'] = self.__dict__['location']
        self.__dataclass_fields__['location'].name = 'in'  # noqa


@dc.dataclass
class MethodExample:
    """
    Method usage example.

    :param params: example parameters
    :param result: example result
    :param name: name for the example pairing
    :param summary: short description for the example pairing
    :param description: a verbose explanation of the example pairing
    """

    params: Dict[str, Any]
    result: Any
    version: str = '2.0'
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class ExampleObject:
    """
    Method usage example.

    :param value: embedded literal example
    :param summary: short description for the example.
    :param description: long description for the example
    :param externalValue: a URL that points to the literal example
    """

    value: MaybeSet[Any] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalValue: MaybeSet[str] = UNSET


@dc.dataclass
class Encoding:
    """
    A single encoding definition applied to a single schema property.

    :param contentType: the Content-Type for encoding a specific property
    :param headers: a map allowing additional information to be provided as headers
    :param style: describes how a specific property value will be serialized depending on its type
    :param explode: when this is true, property values of type array or object generate separate parameters
                    for each value of the array, or key-value-pair of the map
    :param allowReserved: determines whether the parameter value SHOULD allow reserved characters,
                          as defined by RFC3986 to be included without percent-encoding
    """

    contentType: MaybeSet[str] = UNSET
    headers: MaybeSet[Dict[str, Union['Header', Reference]]] = UNSET
    style: MaybeSet[str] = UNSET
    explode: MaybeSet[bool] = UNSET
    allowReserved: MaybeSet[bool] = UNSET


class ParameterLocation(str, enum.Enum):
    """
    The location of the parameter.
    """

    QUERY = 'query'
    HEADER = 'header'
    PATH = 'path'
    COOKIE = 'cookie'


class StyleType(str, enum.Enum):
    """
    Describes how the parameter value will be serialized depending on the type of the parameter value.
    """

    MATRIX = 'matrix'  # path-style parameters defined by RFC6570
    LABEL = 'label'  # label style parameters defined by RFC6570
    FORM = 'form'  # form style parameters defined by RFC6570
    SIMPLE = 'simple'  # simple style parameters defined by RFC6570
    SPACE_DELIMITED = 'spaceDelimited'  # space separated array values
    PIPE_DELIMITED = 'pipeDelimited'  # pipe separated array values
    DEEP_OBJECT = 'deepObject'  # provides a simple way of rendering nested objects using form parameters


@dc.dataclass
class MediaType:
    """
    Each Media Type Object provides schema and examples for the media type identified by its key.

    :param schema: the schema defining the content.
    :param example: example of the media type
    """

    schema: MaybeSet[Dict[str, Any]] = UNSET
    example: MaybeSet[Any] = UNSET
    examples: MaybeSet[Dict[str, ExampleObject]] = UNSET
    encoding: MaybeSet[Dict[str, Encoding]] = UNSET


@dc.dataclass
class RequestBody:
    """
    Describes a single request body.

    :param content: the content of the request body
    :param required: determines if the request body is required in the request
    :param description: a brief description of the request body
    """

    content: Dict[str, MediaType]
    required: MaybeSet[bool] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Parameter:
    """
    Describes a single operation parameter.

    :param name: the name of the parameter
    :param location: the location of the parameter
    :param description: a brief description of the parameter
    :param required: determines whether this parameter is mandatory
    :param deprecated: a parameter is deprecated and SHOULD be transitioned out of usage
    :param allowEmptyValue: the ability to pass empty-valued parameters
    :param style: describes how the parameter value will be serialized depending on the type of the parameter value
    :param explode: when this is true, parameter values of type array or object generate separate parameters
                    for each value of the array or key-value pair of the map
    :param allowReserved: determines whether the parameter value SHOULD allow reserved characters,
                          as defined by RFC3986 :/?#[]@!$&'()*+,;= to be included without percent-encoding
    :param schema: the schema defining the type used for the parameter.
    :param examples: examples of the parameter's potential value
    :param content: a map containing the representations for the parameter
    """

    name: str
    location: ParameterLocation  # `in` field
    description: MaybeSet[str] = UNSET
    required: MaybeSet[bool] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    allowEmptyValue: MaybeSet[bool] = UNSET
    style: MaybeSet[StyleType] = UNSET
    explode: MaybeSet[bool] = UNSET
    allowReserved: MaybeSet[bool] = UNSET
    schema: MaybeSet[JsonSchema] = UNSET
    example: MaybeSet[Any] = UNSET
    examples:  MaybeSet[Dict[str, ExampleObject]] = UNSET
    content: MaybeSet[Dict[str, MediaType]] = UNSET

    def __post_init__(self) -> None:
        # `in` field name is not allowed in python
        self.__dict__['in'] = self.__dict__['location']
        self.__dataclass_fields__['location'].name = 'in'  # noqa


Header = Parameter


@dc.dataclass
class Response:
    """
    A container for the expected responses of an operation.

    :param description: a short description of the response
    :param content: a map containing descriptions of potential response payloads
    """

    description: str
    headers: MaybeSet[Dict[str, Union[Header, Reference]]] = UNSET
    content: MaybeSet[Dict[str, MediaType]] = UNSET
    links: MaybeSet[Dict[str, Union[Link, Reference]]] = UNSET


@dc.dataclass
class Operation:
    """
    Describes a single API operation on a path.

    :param tags: a list of tags for API documentation control
    :param summary: a short summary of what the operation does
    :param description: a verbose explanation of the operation behavior
    :param externalDocs: additional external documentation for this operation
    :param requestBody: the request body applicable for this operation
    :param responses: the list of possible responses as they are returned from executing this operation
    :param deprecated: declares this operation to be deprecated
    :param servers: an alternative server array to service this operation
    :param security: a declaration of which security mechanisms can be used for this operation
    """

    responses: MaybeSet[Dict[str, Union[Response, Reference]]] = UNSET
    requestBody: MaybeSet[Union[RequestBody, Reference]] = UNSET
    tags: MaybeSet[List[str]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    operationId: MaybeSet[str] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    servers: MaybeSet[List[Server]] = UNSET
    security: MaybeSet[List[Dict[str, List[str]]]] = UNSET
    parameters: MaybeSet[List[Union[Parameter, Reference]]] = UNSET


@dc.dataclass
class Path:
    """
    Describes the interface for the given method name.

    :param summary: an optional, string summary, intended to apply to all operations in this path
    :param description: an optional, string description, intended to apply to all operations in this path
    :param servers: an alternative server array to service all operations in this path
    :param parameters: a list of parameters that are applicable for all the operations described under this path
    """

    get: MaybeSet[Operation] = UNSET
    put: MaybeSet[Operation] = UNSET
    post: MaybeSet[Operation] = UNSET
    delete: MaybeSet[Operation] = UNSET
    options: MaybeSet[Operation] = UNSET
    head: MaybeSet[Operation] = UNSET
    patch: MaybeSet[Operation] = UNSET
    trace: MaybeSet[Operation] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    servers: MaybeSet[List[Server]] = UNSET
    parameters: MaybeSet[Union[Parameter, Reference]] = UNSET


@dc.dataclass
class Components:
    """
    Holds a set of reusable objects for different aspects of the OAS.

    :param securitySchemes: an object to hold reusable Security Scheme Objects
    :param schemas: the definition of input and output data types
    """

    schemas: MaybeSet[Dict[str, JsonSchema]] = UNSET
    responses: MaybeSet[Dict[str, Union[Response, Reference]]] = UNSET
    parameters: MaybeSet[Dict[str, Union[Parameter, Reference]]] = UNSET
    examples: MaybeSet[Dict[str, Union[ExampleObject, Reference]]] = UNSET
    requestBodies: MaybeSet[Dict[str, Union[RequestBody, Reference]]] = UNSET
    headers: MaybeSet[Dict[str, Union[Header, Reference]]] = UNSET
    securitySchemes: MaybeSet[Dict[str, Union[SecurityScheme, Reference]]] = UNSET
    links: MaybeSet[Dict[str, Dict[str, Union[Link, Reference]]]] = UNSET
    pathItems: MaybeSet[Dict[str, Union[Path, Reference]]] = UNSET


@dc.dataclass
class SpecRoot:
    """
    The root object of the OpenAPI description.

    :param info: provides metadata about the API
    :param servers: an array of Server Objects, which provide connectivity information to a target server
    :param externalDocs: additional external documentation
    :param openapi: the semantic version number of the OpenAPI Specification version that the OpenAPI document uses
    :param jsonSchemaDialect: the default value for the $schema keyword within Schema Objects
    :param tags: a list of tags used by the specification with additional metadata
    :param security: a declaration of which security mechanisms can be used across the API
    """

    info: Info
    paths: Dict[str, Path]
    components: Components
    servers: MaybeSet[List[Server]] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    tags: MaybeSet[List[Tag]] = UNSET
    security: MaybeSet[List[Dict[str, List[str]]]] = UNSET
    openapi: str = '3.1.0'
    jsonSchemaDialect: MaybeSet[str] = UNSET


class OpenApiMeta(TypedDict):
    params_schema: MaybeSet[Dict[str, JsonSchema]]
    result_schema: MaybeSet[JsonSchema]
    errors: MaybeSet[List[Type[exceptions.JsonRpcError]]]
    examples: MaybeSet[List[MethodExample]]
    tags: MaybeSet[List[Tag]]
    summary: MaybeSet[str]
    description: MaybeSet[str]
    external_docs: MaybeSet[ExternalDocumentation]
    deprecated: MaybeSet[bool]
    security: MaybeSet[List[Dict[str, List[str]]]]
    parameters: MaybeSet[List[Parameter]]
    servers: MaybeSet[List[Server]]
    component_name_prefix: Optional[str]


def annotate(
    params_schema: MaybeSet[Dict[str, JsonSchema]] = UNSET,
    result_schema: MaybeSet[JsonSchema] = UNSET,
    errors: MaybeSet[List[Type[exceptions.JsonRpcError]]] = UNSET,
    examples: MaybeSet[List[MethodExample]] = UNSET,
    tags: MaybeSet[List[Union[str, Tag]]] = UNSET,
    summary: MaybeSet[str] = UNSET,
    description: MaybeSet[str] = UNSET,
    external_docs: MaybeSet[ExternalDocumentation] = UNSET,
    deprecated: MaybeSet[bool] = UNSET,
    security: MaybeSet[List[Dict[str, List[str]]]] = UNSET,
    parameters: MaybeSet[List[Parameter]] = UNSET,
    servers: MaybeSet[List[Server]] = UNSET,
    component_name_prefix: Optional[str] = None,
) -> Callable[[Func], Func]:
    """
    Adds Open Api specification annotation to the method.

    :param params_schema: method parameters JSON schema
    :param result_schema: method result JSON schema
    :param errors: method errors
    :param examples: method usage examples
    :param tags: a list of tags for method documentation control
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param external_docs: an external resource for extended documentation
    :param deprecated: declares this method to be deprecated
    :param security: a declaration of which security mechanisms can be used for the method
    :param parameters: a list of parameters that are applicable for the method
    :param servers: a list of connectivity information of a target server.
    :param component_name_prefix: components name prefix.
    """

    def decorator(method: Func) -> Func:
        meta: OpenApiMeta = dict(
            params_schema=params_schema,
            result_schema=result_schema,
            errors=errors,
            examples=examples,
            tags=[
                tag if isinstance(tag, Tag) else Tag(name=tag) for tag in tags
            ] if not isinstance(tags, UnsetType) else UNSET,
            summary=summary,
            description=description,
            external_docs=external_docs,
            deprecated=deprecated,
            security=security,
            parameters=parameters,
            servers=servers,
            component_name_prefix=component_name_prefix,
        )
        utils.set_meta(method, openapi_spec=meta)

        return method

    return decorator


class OpenAPI(Specification):
    """
    OpenAPI Specification.

    :param info: provides metadata about the API
    :param servers: an array of Server Objects, which provide connectivity information to a target server
    :param external_docs: additional external documentation
    :param openapi: the semantic version number of the OpenAPI Specification version that the OpenAPI document uses
    :param json_schema_dialect: the default value for the $schema keyword within Schema Objects
    :param tags: a list of tags used by the specification with additional metadata
    :param security: a declaration of which security mechanisms can be used across the API
    :param schema_extractor: method specification extractor
    :param schema_extractors: method specification extractors. Extractors results will be merged in reverse order
                              (former extractor rewrites the result of the later one)
    :param path: specification url path
    :param security_schemes: an object to hold reusable Security Scheme Objects
    :param ui: web ui instance
    :param ui_path: wet ui path
    """

    def __init__(
        self,
        info: Info,
        path: str = '/openapi.json',
        servers: MaybeSet[List[Server]] = UNSET,
        external_docs: MaybeSet[ExternalDocumentation] = UNSET,
        tags: MaybeSet[List[Tag]] = UNSET,
        security: MaybeSet[List[Dict[str, List[str]]]] = UNSET,
        security_schemes: MaybeSet[Dict[str, Union[SecurityScheme, Reference]]] = UNSET,
        openapi: str = '3.1.0',
        json_schema_dialect: MaybeSet[str] = UNSET,
        schema_extractor: Optional[extractors.BaseSchemaExtractor] = None,
        schema_extractors: Iterable[extractors.BaseSchemaExtractor] = (),
        ui: Optional[BaseUI] = None,
        ui_path: str = '/ui/',
        error_http_status_map: Dict[int, int] = {},
    ):
        super().__init__(path, ui=ui, ui_path=ui_path)

        self._spec = SpecRoot(
            info=info,
            paths={},
            components=Components(securitySchemes=security_schemes),
            servers=servers,
            externalDocs=external_docs,
            tags=tags,
            security=security,
            openapi=openapi,
            jsonSchemaDialect=json_schema_dialect,
        )
        self._error_http_status_map = error_http_status_map
        self._schema_extractors = list(schema_extractors) or [schema_extractor or extractors.BaseSchemaExtractor()]

    def schema(
        self,
        path: str,
        methods_map: Mapping[str, Iterable[Method]] = {},
        component_name_prefix: str = '',
    ) -> Dict[str, Any]:
        spec = copy.deepcopy(self._spec)

        methods_list = [
            (utils.join_path(path, prefix), method)
            for prefix, methods in methods_map.items()
            for method in methods
        ]

        for prefix, method in methods_list:
            method_meta = utils.get_meta(method.method)
            annotated_spec: OpenApiMeta = method_meta.get('openapi_spec', {})

            component_name_prefix = annotated_spec.get('component_name_prefix') or component_name_prefix
            status_errors_map = self._extract_errors(method)
            default_status_errors = status_errors_map.pop(HTTP_DEFAULT_STATUS, [])

            errors_schema = self._extract_errors_schema(spec, method, status_errors_map, component_name_prefix)

            request_schema = self._extract_request_schema(spec, method, component_name_prefix)
            response_schema = self._extract_response_schema(spec, method, default_status_errors, component_name_prefix)

            summary, description = self._extract_description(method)
            tags = self._extract_tags(method)
            servers = self._extract_servers(method)
            parameters = self._extract_parameters(method)
            security = self._extract_security(method)
            deprecated = self._extract_deprecated(method)
            external_docs = self._extract_external_docs(method)

            request_examples, response_success_examples = self._build_examples(
                method, annotated_spec.get('examples', UNSET) or [],
            )

            spec.paths[f'{prefix}#{method.name}'] = Path(
                post=Operation(
                    requestBody=RequestBody(
                        description='JSON-RPC Request',
                        content={
                            JSONRPC_MEDIATYPE: MediaType(
                                schema=request_schema,
                                examples=request_examples or UNSET,
                            ),
                        },
                        required=True,
                    ),
                    responses={
                        **{
                            str(HTTP_DEFAULT_STATUS): Response(
                                description=(response_schema or {}).get('description', 'JSON-RPC Response'),
                                content={
                                    JSONRPC_MEDIATYPE: MediaType(
                                        schema=response_schema,
                                        examples=response_success_examples or UNSET,
                                    ),
                                },
                            ),
                        },
                        **{
                            str(status): Response(
                                description=error_schema.get('description', 'JSON-RPC Error'),
                                content={
                                    JSONRPC_MEDIATYPE: MediaType(
                                        schema=error_schema,
                                    ),
                                },
                            )
                            for status, error_schema in errors_schema.items()
                        },
                    },
                    tags=[tag.name for tag in tags] or UNSET,
                    summary=summary,
                    description=description,
                    deprecated=deprecated,
                    externalDocs=external_docs,
                    security=security or UNSET,
                    parameters=list(parameters) or UNSET,
                    servers=servers or UNSET,
                ),
            )

        return drop_unset(dc.asdict(spec))

    def _extract_errors(self, method: Method) -> Dict[int, List[Type[exceptions.JsonRpcError]]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        errors = annotations.get('errors', UNSET) or []
        for schema_extractor in self._schema_extractors:
            errors.extend(schema_extractor.extract_errors(method.method) or [])

        unique_errors = list({error.code: error for error in errors}.values())
        status_error_map: Dict[int, List[Type[exceptions.JsonRpcError]]] = defaultdict(list)
        for error in unique_errors:
            http_status = self._error_http_status_map.get(error.code, HTTP_DEFAULT_STATUS)
            status_error_map[http_status].append(error)

        return status_error_map

    def _extract_errors_schema(
            self,
            spec: SpecRoot,
            method: Method,
            status_errors_map: Dict[int, List[Type[exceptions.JsonRpcError]]],
            component_name_prefix: str,
    ) -> Dict[int, Dict[str, Any]]:
        status_error_schema_map: Dict[int, Dict[str, Any]] = {}

        for status, errors in status_errors_map.items():
            for schema_extractor in self._schema_extractors:
                if result := schema_extractor.extract_error_response_schema(
                    method.name,
                    method.method,
                    ref_template=f'#/components/schemas/{component_name_prefix}{{model}}',
                    errors=errors,
                ):
                    schema, components = result
                    if components:
                        spec.components.schemas = schemas = spec.components.schemas or {}
                        schemas.update({
                            f"{component_name_prefix}{name}": component
                            for name, component in components.items()
                        })
                    status_error_schema_map[status] = schema
                    break

        return status_error_schema_map

    def _extract_request_schema(
            self,
            spec: SpecRoot,
            method: Method,
            component_name_prefix: str,
    ) -> MaybeSet[Dict[str, Any]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        request_schema: MaybeSet[Dict[str, Any]] = UNSET
        if params_schema := annotations.get('params_schema', UNSET):
            request_schema = build_request_schema(method.name, params_schema)
        else:
            for schema_extractor in self._schema_extractors:
                if result := schema_extractor.extract_request_schema(
                    method.name,
                    method.method,
                    ref_template=f'#/components/schemas/{component_name_prefix}{{model}}',
                    exclude=[method.context] if method.context else [],
                ):
                    request_schema, components = result
                    if components:
                        spec.components.schemas = schemas = spec.components.schemas or {}
                        schemas.update({
                            f"{component_name_prefix}{name}": component
                            for name, component in components.items()
                        })
                    break

        return request_schema

    def _extract_response_schema(
            self,
            spec: SpecRoot,
            method: Method,
            errors: List[Type[exceptions.JsonRpcError]],
            component_name_prefix: str,
    ) -> MaybeSet[Dict[str, Any]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        response_schema: MaybeSet[Dict[str, Any]] = UNSET
        if result_schema := annotations.get('result_schema', UNSET):
            response_schema = build_response_schema(result_schema, errors=errors)
        else:
            for schema_extractor in self._schema_extractors:
                if result := schema_extractor.extract_response_schema(
                    method.name,
                    method.method,
                    ref_template=f'#/components/schemas/{component_name_prefix}{{model}}',
                    errors=errors,
                ):
                    response_schema, components = result
                    if components:
                        spec.components.schemas = schemas = spec.components.schemas or {}
                        schemas.update({
                            f"{component_name_prefix}{name}": component
                            for name, component in components.items()
                        })
                    break

        return response_schema

    def _extract_description(self, method: Method) -> Tuple[MaybeSet[str], MaybeSet[str]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        summary = annotations.get('summary', UNSET)
        description = annotations.get('description', UNSET)

        for schema_extractor in self._schema_extractors:
            if not summary:
                summary = schema_extractor.extract_summary(method.method)
            if not description:
                description = schema_extractor.extract_description(method.method)

        return summary, description

    def _extract_tags(self, method: Method) -> List[Tag]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        tags = annotations.get('tags', UNSET) or []

        return tags

    def _extract_servers(self, method: Method) -> List[Server]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        servers = annotations.get('servers', UNSET) or []

        return servers

    def _extract_parameters(self, method: Method) -> List[Parameter]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        parameters = annotations.get('parameters', UNSET) or []

        return parameters

    def _extract_security(self, method: Method) -> List[Dict[str, List[str]]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        security = annotations.get('security', UNSET) or []

        return security

    def _extract_deprecated(self, method: Method) -> MaybeSet[bool]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        deprecated = annotations.get('deprecated', UNSET)

        for schema_extractor in self._schema_extractors:
            deprecated = deprecated or schema_extractor.extract_deprecation_status(method.method)

        return deprecated

    def _extract_external_docs(self, method: Method) -> MaybeSet[ExternalDocumentation]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenApiMeta = method_meta.get('openapi_spec', {})

        external_docs = annotations.get('external_docs', UNSET)

        return external_docs

    def _build_examples(self, method: Method, examples: List[MethodExample]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        request_examples: Dict[str, Any] = {
            example.summary or f'Example#{i}': ExampleObject(
                summary=example.summary,
                description=example.description,
                value={
                    'jsonrpc': example.version,
                    'id': 1,
                    'method': method.name,
                    'params': example.params,
                },
            ) for i, example in enumerate(examples)
        }

        response_examples: Dict[str, Any] = {
            example.summary or f'Example#{i}': ExampleObject(
                summary=example.summary,
                description=example.description,
                value={
                    'jsonrpc': example.version,
                    'id': 1,
                    'result': example.result,
                },
            ) for i, example in enumerate(examples)
        }

        return request_examples, response_examples


class SwaggerUI(BaseUI):
    """
    Swagger UI.

    :param config: documentation configurations
                   (see https://github.com/swagger-api/swagger-ui/blob/master/docs/usage/configuration.md).
    """

    def __init__(self, **configs: Any):
        try:
            import openapi_ui_bundles
        except ImportError:
            raise RuntimeError(
                "openapi-ui-bundles package not found. "
                "Please install pjrpc with extra requirement: pip install pjrpc[openapi-ui-bundles]",
            )

        self._bundle = openapi_ui_bundles
        self._configs = configs

    def get_static_folder(self) -> str:
        return str(self._bundle.swagger_ui.static_path)

    @ft.lru_cache(maxsize=10)
    def get_index_page(self, spec_url: str) -> str:
        index_path = pathlib.Path(self.get_static_folder()) / 'index.html'
        index_page = index_path.read_text()

        config = dict(self._configs, **{'url': spec_url, 'dom_id': '#swagger-ui'})
        config_str = ', '.join(f'{param}: "{value}"' for param, value in config.items())

        return re.sub(
            pattern=r'SwaggerUIBundle\({.*?}\)',
            repl=f'SwaggerUIBundle({{ {config_str} }})',
            string=index_page,
            count=1,
            flags=re.DOTALL,
        )


class RapiDoc(BaseUI):
    """
    RapiDoc UI.

    :param config: documentation configurations (see https://mrin9.github.io/RapiDoc/api.html).
                   Be aware that configuration parameters should be in snake case,
                   for example: parameter `heading-text` should be passed as `heading_text`)
    """

    def __init__(self, **configs: Any):
        try:
            import openapi_ui_bundles.rapidoc
        except ImportError:
            raise RuntimeError(
                "openapi-ui-bundles package not found. "
                "Please install pjrpc with extra requirement: pip install pjrpc[openapi-ui-bundles]",
            )

        self._bundle = openapi_ui_bundles.rapidoc
        self._configs = configs

    def get_static_folder(self) -> str:
        return str(self._bundle.static_path)

    @ft.lru_cache(maxsize=10)
    def get_index_page(self, spec_url: str) -> str:
        index_path = pathlib.Path(self.get_static_folder()) / 'index.html'
        index_page = index_path.read_text()

        config = dict(self._configs, **{'spec_url': spec_url, 'id': 'thedoc'})
        config_str = ' '.join(f'{param.replace("_", "-")}="{value}"' for param, value in config.items())

        return re.sub(
            pattern='<rapi-doc.*?>',
            repl=f'<rapi-doc {config_str}>',
            string=index_page,
            count=1,
            flags=re.DOTALL,
        )


class ReDoc(BaseUI):
    """
    ReDoc UI.

    :param config: documentation configurations (see https://github.com/Redocly/redoc#configuration).
                   Be aware that configuration parameters should be in snake case,
                   for example: parameter `heading-text` should be passed as `heading_text`)
    """

    def __init__(self, **configs: Any):
        try:
            import openapi_ui_bundles.redoc
        except ImportError:
            raise RuntimeError(
                "openapi-ui-bundles package not found. "
                "Please install pjrpc with extra requirement: pip install pjrpc[openapi-ui-bundles]",
            )

        self._bundle = openapi_ui_bundles.redoc
        self._configs = configs

    def get_static_folder(self) -> str:
        return str(self._bundle.static_path)

    @ft.lru_cache(maxsize=10)
    def get_index_page(self, spec_url: str) -> str:
        index_path = pathlib.Path(self.get_static_folder()) / 'index.html'
        index_page = index_path.read_text()

        config = dict(self._configs, **{'spec_url': spec_url})
        config_str = ' '.join(f'{param.replace("_", "-")}="{value}"' for param, value in config.items())

        return re.sub(
            pattern='<redoc.*?>',
            repl=f'<redoc {config_str}>',
            string=index_page,
            count=1,
            flags=re.DOTALL,
        )



from typing import Any, Dict, Optional

from aiohttp import client

import pjrpc
from pjrpc.client import AbstractAsyncClient


class Client(AbstractAsyncClient):
    """
    `Aiohttp <https://aiohttp.readthedocs.io/en/stable/client.html>`_ library client backend.

    :param url: url to be used as JSON-RPC endpoint
    :param session_args: additional :py:class:`aiohttp.ClientSession` arguments
    :param session: custom session to be used instead of :py:class:`aiohttp.ClientSession`
    """

    def __init__(
        self, url: str,
        session_args: Optional[Dict[str, Any]] = None,
        session: Optional[client.ClientSession] = None,
        raise_for_status: bool = True,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._endpoint = url
        self._session = session or client.ClientSession(**(session_args or {}))
        self._raise_for_status = raise_for_status

    async def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param data: request text
        :param is_notification: is the request a notification
        :returns: response text
        """

        headers = kwargs.setdefault('headers', {})
        headers['Content-Type'] = pjrpc.common.DEFAULT_CONTENT_TYPE

        async with self._session.post(self._endpoint, data=request_text, **kwargs) as resp:
            if self._raise_for_status:
                resp.raise_for_status()
            response_text = await resp.text()

        if is_notification:
            return None

        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in pjrpc.common.RESPONSE_CONTENT_TYPES:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    async def close(self) -> None:
        """
        Closes current http session.
        """

        await self._session.close()

    async def __aenter__(self) -> 'Client':
        await self._session.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self._session.__aexit__(*args)



import functools as ft
import inspect
from typing import Any, Callable, Dict, Iterable, List, Optional

import pydantic

from pjrpc.common.typedefs import JsonRpcParams
from pjrpc.server.typedefs import ExcludeFunc

from . import base


class PydanticValidator(base.BaseValidator):
    """
    Parameters validator based on `pydantic <https://pydantic-docs.helpmanual.io/>`_ library.
    Uses python type annotations for parameters validation.

    :param coerce: if ``True`` returns converted (coerced) parameters according to parameter type annotation
                   otherwise returns parameters as is
    :param exclude_param: a function that decides if the parameters must be excluded
                          from validation (useful for dependency injection)
    """

    def __init__(self, coerce: bool = True, exclude_param: Optional[ExcludeFunc] = None, **config_args: Any):
        super().__init__(exclude_param=exclude_param)
        self._coerce = coerce

        config_args.setdefault('extra', 'forbid')

        # https://pydantic-docs.helpmanual.io/usage/model_config/
        self._model_config = pydantic.ConfigDict(**config_args)  # type: ignore[typeddict-item]

    def validate_method(
        self, method: Callable[..., Any], params: Optional['JsonRpcParams'], exclude: Iterable[str] = (), **kwargs: Any,
    ) -> Dict[str, Any]:
        """
        Validates params against method using ``pydantic`` validator.

        :param method: method to validate parameters against
        :param params: parameters to be validated
        :param exclude: parameter names to be excluded from validation

        :returns: coerced parameters if `coerce` flag is ``True`` otherwise parameters as is
        :raises: ValidationError
        """

        signature = self.signature(method, tuple(exclude))
        schema = self.build_validation_schema(signature)

        params_model = pydantic.create_model(method.__name__, **schema, model_config=self._model_config)

        bound_params = self.bind(signature, params)
        try:
            obj = params_model(**bound_params.arguments)
        except pydantic.ValidationError as e:
            raise base.ValidationError(*e.errors()) from e

        return {attr: getattr(obj, attr) for attr in obj.model_fields} if self._coerce else bound_params.arguments

    @ft.lru_cache(maxsize=None)
    def build_validation_schema(self, signature: inspect.Signature) -> Dict[str, Any]:
        """
        Builds pydantic model based validation schema from method signature.

        :param signature: method signature to build schema for
        :returns: validation schema
        """

        field_definitions = {}

        for param in signature.parameters.values():
            if param.kind is inspect.Parameter.VAR_KEYWORD:
                field_definitions[param.name] = (
                    Optional[Dict[str, param.annotation]]  # type: ignore
                    if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else None,
                )
            elif param.kind is inspect.Parameter.VAR_POSITIONAL:
                field_definitions[param.name] = (
                    Optional[List[param.annotation]]  # type: ignore
                    if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else None,
                )
            else:
                field_definitions[param.name] = (
                    param.annotation if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else ...,
                )

        return field_definitions



import logging
from typing import Any, Dict, Optional

import aio_pika
from yarl import URL

import pjrpc.server

logger = logging.getLogger(__package__)

AioPikaDispatcher = pjrpc.server.AsyncDispatcher[aio_pika.abc.AbstractIncomingMessage]


class Executor:
    """
    `aio_pika <https://aio-pika.readthedocs.io/en/latest/>`_ based JSON-RPC server.

    :param broker_url: broker connection url
    :param rx_queue_name: requests queue name
    :param tx_exchange_name: response exchange name
    :param tx_routing_key: response routing key
    :param prefetch_count: worker prefetch count
    :param kwargs: dispatcher additional arguments
    """

    def __init__(
        self,
        broker_url: URL,
        rx_queue_name: str,
        tx_exchange_name: Optional[str] = None,
        tx_routing_key: Optional[str] = None,
        prefetch_count: int = 0,
        **kwargs: Any,
    ):
        self._broker_url = broker_url
        self._rx_queue_name = rx_queue_name
        self._tx_exchange_name = tx_exchange_name
        self._tx_routing_key = tx_routing_key
        self._prefetch_count = prefetch_count

        self._connection = aio_pika.connection.Connection(broker_url)
        self._channel: Optional[aio_pika.abc.AbstractChannel] = None

        self._queue: Optional[aio_pika.abc.AbstractQueue] = None
        self._exchange: Optional[aio_pika.abc.AbstractExchange] = None
        self._consumer_tag: Optional[str] = None

        self._dispatcher = AioPikaDispatcher(**kwargs)

    @property
    def dispatcher(self) -> AioPikaDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def start(
            self,
            queue_args: Optional[Dict[str, Any]] = None,
            exchange_args: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Starts executor.

        :param queue_args: queue arguments
        :param exchange_args: exchange arguments
        """

        await self._connection.connect()
        self._channel = channel = await self._connection.channel()

        self._queue = queue = await channel.declare_queue(self._rx_queue_name, **(queue_args or {}))
        if self._tx_exchange_name:
            self._exchange = await channel.declare_exchange(self._tx_exchange_name, **(exchange_args or {}))
        await channel.set_qos(prefetch_count=self._prefetch_count)
        self._consumer_tag = await queue.consume(self._rpc_handle)

    async def shutdown(self) -> None:
        """
        Stops executor.
        """

        if self._consumer_tag and self._queue:
            await self._queue.cancel(self._consumer_tag)
        if self._channel:
            await self._channel.close()

        await self._connection.close()

    async def _rpc_handle(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        """
        Handles JSON-RPC request.

        :param message: incoming message
        """

        try:
            reply_to = message.reply_to
            response = await self._dispatcher.dispatch(message.body.decode(), context=message)

            if response is not None:
                response_text, error_codes = response
                if self._tx_routing_key:
                    routing_key = self._tx_routing_key
                elif reply_to:
                    routing_key = reply_to
                else:
                    routing_key = ""
                    logger.warning("property 'reply_to' or 'tx_routing_key' missing")
                async with self._connection.channel() as channel:
                    exchange = self._exchange if self._exchange else channel.default_exchange
                    await exchange.publish(
                        aio_pika.Message(
                            body=response_text.encode(),
                            reply_to=reply_to,
                            correlation_id=message.correlation_id,
                            content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                        ),
                        routing_key=routing_key,
                    )

            await message.ack()

        except Exception as e:
            logger.exception("jsonrpc request handling error: %s", e)



from typing import Any, List, Optional

import httpx

import pjrpc
from pjrpc.client import AbstractAsyncClient, AbstractClient


class Client(AbstractClient):
    """
    `httpx <https://www.python-httpx.org/>`_ library sync client backend.

    :param url: url to be used as JSON-RPC endpoint.
    :param session: custom session to be used instead of :py:class:`requests.Session`
    :param kwargs: parameters to be passed to :py:class:`pjrpc.client.AbstractClient`
    """

    def __init__(self, url: str, client: Optional[httpx.Client] = None, raise_for_status: bool = True, **kwargs: Any):
        super().__init__(**kwargs)
        self._endpoint = url
        self._client = client or httpx.Client()
        self._raise_for_status = raise_for_status

    def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param data: request text
        :param is_notification: is the request a notification
        :returns: response text
        """

        headers = kwargs.setdefault('headers', {})
        headers['Content-Type'] = pjrpc.common.DEFAULT_CONTENT_TYPE

        resp = self._client.post(self._endpoint, content=request_text, **kwargs)
        if self._raise_for_status:
            resp.raise_for_status()
        if is_notification:
            return None

        response_text = resp.text
        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in pjrpc.common.RESPONSE_CONTENT_TYPES:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    def close(self) -> None:
        """
        Closes the current http session.
        """

        self._client.close()

    def __enter__(self) -> 'Client':
        self._client.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        self._client.__exit__(*args)


class AsyncClient(AbstractAsyncClient):
    """
    `httpx <https://www.python-httpx.org/>`_ library async client backend.

    :param url: url to be used as JSON-RPC endpoint
    :param session_args: additional :py:class:`aiohttp.ClientSession` arguments
    :param session: custom session to be used instead of :py:class:`aiohttp.ClientSession`
    """

    def __init__(
        self,
        url: str,
        client: Optional[httpx.AsyncClient] = None,
        raise_for_status: bool = True,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._endpoint = url
        self._client = client or httpx.AsyncClient()
        self._raise_for_status = raise_for_status

    async def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param data: request text
        :param is_notification: is the request a notification
        :returns: response text
        """

        headers = kwargs.setdefault('headers', {})
        headers['Content-Type'] = pjrpc.common.DEFAULT_CONTENT_TYPE

        resp = await self._client.post(self._endpoint, content=request_text, **kwargs)
        if self._raise_for_status:
            resp.raise_for_status()

        response_buff: List[str] = []
        async for chunk in resp.aiter_text():
            response_buff.append(chunk)

        response_text = ''.join(response_buff)
        if is_notification:
            return None

        content_type = resp.headers.get('Content-Type', '')
        if response_text and content_type.split(';')[0] not in pjrpc.common.RESPONSE_CONTENT_TYPES:
            raise pjrpc.exc.DeserializationError(f"unexpected response content type: {content_type}")

        return response_text

    async def close(self) -> None:
        """
        Closes current http session.
        """

        await self._client.aclose()

    async def __aenter__(self) -> 'AsyncClient':
        await self._client.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self._client.__aexit__(*args)



import logging
from typing import Any, Dict, Optional, Union

import kombu.mixins

import pjrpc
from pjrpc.client import AbstractClient
from pjrpc.common import UNSET, MaybeSet, UnsetType

logger = logging.getLogger(__package__)


class Client(AbstractClient):
    """
    `kombu <https://aio-pika.readthedocs.io/en/latest/>`_ based JSON-RPC client.
    Note: the client is not thread-safe.

    :param broker_url: broker connection url
    :param conn_args: broker connection arguments.
    :param queue_name: queue name to publish requests to
    :param exchange_name: exchange to publish requests to. If ``None`` default exchange is used
    :param exchange_args: exchange arguments
    :param routing_key: reply message routing key. If ``None`` queue name is used
    :param result_queue_name: result queue name. If ``None`` random exclusive queue is declared for each request
    :param conn_args: additional connection arguments
    :param kwargs: parameters to be passed to :py:class:`pjrpc.client.AbstractClient`
    """

    def __init__(
        self,
        broker_url: str,
        queue_name: Optional[str] = None,
        conn_args: Optional[Dict[str, Any]] = None,
        exchange_name: Optional[str] = None,
        exchange_args: Optional[Dict[str, Any]] = None,
        routing_key: Optional[str] = None,
        result_queue_name: Optional[str] = None,
        result_queue_args: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        assert queue_name or routing_key, "queue_name or routing_key must be provided"

        super().__init__(**kwargs)
        self._connection = kombu.Connection(broker_url, **(conn_args or {}))
        self._routing_key = routing_key or queue_name
        self._result_queue = None
        self._result_queue_args = result_queue_args
        self._exchange = None
        self._exchange_args = exchange_args

        if exchange_name:
            self._exchange = kombu.Exchange(exchange_name,  **(exchange_args or {}))
        if result_queue_name:
            self._result_queue = kombu.Queue(result_queue_name, **(result_queue_args or {}))

    def close(self) -> None:
        """
        Closes the current broker connection.
        """

        self._connection.close()

    def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Sends a JSON-RPC request.

        :param request_text: request text
        :param is_notification: is the request a notification
        :param kwargs: publish additional arguments
        :returns: response text
        """

        if is_notification:
            with kombu.Producer(self._connection) as producer:
                producer.publish(
                    request_text,
                    exchange=self._exchange or '',
                    routing_key=self._routing_key,
                    content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                    **kwargs,
                )
                return None

        request_id = kombu.uuid()
        result_queue = self._result_queue or kombu.Queue(
            exclusive=True, name=request_id, **(self._result_queue_args or {}),
        )

        with kombu.Producer(self._connection) as producer:
            result_queue.declare(channel=self._connection.default_channel)
            producer.publish(
                request_text,
                exchange=self._exchange or '',
                routing_key=self._routing_key,
                reply_to=result_queue.name,
                correlation_id=request_id,
                content_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
                **kwargs,
            )

        response: MaybeSet[Union[None, str, Exception]] = UNSET

        def on_response(message: kombu.Message) -> None:
            nonlocal response

            try:
                if message.properties.get('correlation_id') != request_id:
                    logger.warning("unexpected message received: %r", message)
                    return

                if message.content_type not in pjrpc.common.RESPONSE_CONTENT_TYPES:
                    raise pjrpc.exc.DeserializationError(f"unexpected response content type: {message.content_type}")
                else:
                    response = message.body
            except Exception as e:
                response = e

        with kombu.Consumer(self._connection, on_message=on_response, queues=result_queue, no_ack=True):
            while response is UNSET:
                self._connection.drain_events(timeout=kwargs.get('timeout', None))

        if isinstance(response, Exception):
            raise response

        assert not isinstance(response, UnsetType), "response is unset"
        return response



"""
aiohttp JSON-RPC server integration.
"""

import functools as ft
import json
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Tuple

from starlette import exceptions, routing
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.staticfiles import StaticFiles

import pjrpc
from pjrpc.server import specs, utils

StarletteDispatcher = pjrpc.server.AsyncDispatcher[Request]


class Application:
    """
    `starlette <https://www.starlette.io/>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param spec: api specification instance
    :param app: starlette application instance
    :param status_by_error: a function returns http status code by json-rpc error codes, 200 for all errors by default
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(
        self,
        path: str = '',
        spec: Optional[specs.Specification] = None,
        specs: Iterable[specs.Specification] = (),
        app: Optional[Starlette] = None,
        status_by_error: Callable[[Tuple[int, ...]], int] = lambda codes: 200,
        **kwargs: Any,
    ):
        self._path = path = path.rstrip('/')
        self._specs = ([spec] if spec else []) + list(specs)
        self._app = app or Starlette()
        self._status_by_error = status_by_error
        self._dispatcher = StarletteDispatcher(**kwargs)
        self._endpoints: Dict[str, StarletteDispatcher] = {'': self._dispatcher}

        self._app.add_route(path, ft.partial(self._rpc_handle, dispatcher=self._dispatcher), methods=['POST'])

        for spec in self._specs:
            self._app.add_route(
                utils.join_path(path, spec.path),
                ft.partial(self._generate_spec, spec=spec),
                methods=['GET'],
            )

            if spec.ui and spec.ui_path:
                ui_path = utils.join_path(path, spec.ui_path)

                self._app.add_route(
                    utils.join_path(ui_path, '/'), ft.partial(self._ui_index_page, spec=spec),
                )
                self._app.add_route(
                    utils.join_path(ui_path, 'index.html'), ft.partial(self._ui_index_page, spec=spec),
                )
                self._app.routes.append(
                    routing.Mount(ui_path, app=StaticFiles(directory=str(spec.ui.get_static_folder()))),
                )

    @property
    def app(self) -> Starlette:
        """
        aiohttp application.
        """

        return self._app

    @property
    def dispatcher(self) -> StarletteDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    @property
    def endpoints(self) -> Mapping[str, StarletteDispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_endpoint(self, prefix: str, **kwargs: Any) -> StarletteDispatcher:
        """
        Adds additional endpoint.

        :param prefix: endpoint prefix
        :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
        :return: dispatcher
        """

        dispatcher = StarletteDispatcher(**kwargs)
        self._endpoints[prefix] = dispatcher

        self._app.add_route(
            utils.join_path(self._path, prefix),
            ft.partial(self._rpc_handle, dispatcher=dispatcher),
            methods=['POST'],
        )

        return dispatcher

    def generate_spec(self, spec: specs.Specification, path: str = '') -> Dict[str, Any]:
        methods = {path: dispatcher.registry.values() for path, dispatcher in self._endpoints.items()}
        return spec.schema(path=path, methods_map=methods)

    async def _generate_spec(self, request: Request, spec: specs.Specification) -> Response:
        endpoint_path = utils.remove_suffix(request.url.path, suffix=spec.path)
        schema = self.generate_spec(path=endpoint_path, spec=spec)

        return Response(
            content=json.dumps(schema, cls=specs.JSONEncoder),
            media_type='application/json',
        )

    async def _ui_index_page(self, request: Request, spec: specs.Specification) -> Response:
        assert spec.ui is not None, "spec is not set"

        app_path = request.url.path.rsplit(spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, spec.path)

        return Response(
            content=spec.ui.get_index_page(spec_url=spec_full_path),
            media_type='text/html',
        )

    async def _rpc_handle(self, http_request: Request, dispatcher: StarletteDispatcher) -> Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.headers['Content-Type'] not in pjrpc.common.REQUEST_CONTENT_TYPES:
            raise exceptions.HTTPException(415)

        try:
            request_data = await http_request.body()
            request_text = request_data.decode()
        except UnicodeDecodeError as e:
            raise exceptions.HTTPException(400) from e

        response = await dispatcher.dispatch(request_text, context=http_request)
        if response is None:
            return Response()
        else:
            response_text, error_codes = response
            return Response(
                status_code=self._status_by_error(error_codes),
                content=response_text,
                media_type=pjrpc.common.DEFAULT_CONTENT_TYPE,
            )



"""
aiohttp JSON-RPC server integration.
"""

import functools as ft
import json
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Tuple

import aiohttp.web
from aiohttp import web

import pjrpc
from pjrpc.server import specs, utils

AioHttpDispatcher = pjrpc.server.AsyncDispatcher[web.Request]


class Application:
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param spec: api specification instance
    :param app: aiohttp application instance
    :param status_by_error: a function returns http status code by json-rpc error codes, 200 for all errors by default
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(
        self,
        path: str = '',
        spec: Optional[specs.Specification] = None,
        specs: Iterable[specs.Specification] = (),
        app: Optional[web.Application] = None,
        status_by_error: Callable[[Tuple[int, ...]], int] = lambda codes: 200,
        **kwargs: Any,
    ):
        self._path = path = path.rstrip('/')
        self._specs = ([spec] if spec else []) + list(specs)
        self._app = app or web.Application()
        self._status_by_error = status_by_error
        self._dispatcher = AioHttpDispatcher(**kwargs)
        self._endpoints: Dict[str, AioHttpDispatcher] = {'': self._dispatcher}

        self._app.router.add_post(path, ft.partial(self._rpc_handle, dispatcher=self._dispatcher))

        for spec in self._specs:
            self._app.router.add_get(
                utils.join_path(path, spec.path),
                ft.partial(self._generate_spec, spec=spec),
            )

            if spec.ui and spec.ui_path:
                ui_app = web.Application()
                ui_app.router.add_get('/', ft.partial(self._ui_index_page, spec=spec))
                ui_app.router.add_get('/index.html', ft.partial(self._ui_index_page, spec=spec))
                ui_app.router.add_static('/', spec.ui.get_static_folder())

                self._app.add_subapp(utils.join_path(path, spec.ui_path), ui_app)

    @property
    def app(self) -> web.Application:
        """
        aiohttp application.
        """

        return self._app

    @property
    def dispatcher(self) -> AioHttpDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    @property
    def endpoints(self) -> Mapping[str, AioHttpDispatcher]:
        """
        JSON-RPC application registered endpoints.
        """

        return self._endpoints

    def add_subapp(self, prefix: str, subapp: 'Application') -> None:
        """
        Adds sub-application.

        :param prefix: sub-application prefix
        :param subapp: sub-application to be added
        """

        prefix = prefix.rstrip('/')
        self._endpoints[utils.join_path(prefix, subapp._path)] = subapp.dispatcher
        self._app.add_subapp(utils.join_path(self._path, prefix), subapp.app)

    def add_endpoint(
        self,
        prefix: str,
        subapp: Optional[aiohttp.web.Application] = None,
        **kwargs: Any,
    ) -> AioHttpDispatcher:
        """
        Adds additional endpoint.

        :param prefix: endpoint prefix
        :param subapp: aiohttp subapp the endpoint will be served on
        :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
        :return: dispatcher
        """

        prefix = prefix.rstrip('/')
        dispatcher = AioHttpDispatcher(**kwargs)
        self._endpoints[prefix] = dispatcher

        if subapp:
            subapp.router.add_post('', ft.partial(self._rpc_handle, dispatcher=dispatcher))
            self._app.add_subapp(utils.join_path(self._path, prefix), subapp)
        else:
            self._app.router.add_post(
                utils.join_path(self._path, prefix),
                ft.partial(self._rpc_handle, dispatcher=dispatcher),
            )

        return dispatcher

    def generate_spec(self, spec: specs.Specification, path: str = '') -> Dict[str, Any]:
        methods = {path: dispatcher.registry.values() for path, dispatcher in self._endpoints.items()}
        return spec.schema(path=path, methods_map=methods)

    async def _generate_spec(self, request: web.Request, spec: specs.Specification) -> web.Response:
        endpoint_path = utils.remove_suffix(request.path, suffix=spec.path)
        schema = self.generate_spec(path=endpoint_path, spec=spec)

        return web.json_response(text=json.dumps(schema, cls=specs.JSONEncoder))

    async def _ui_index_page(self, request: web.Request, spec: specs.Specification) -> web.Response:
        assert spec.ui is not None, "spec is not set"

        app_path = request.path.rsplit(spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, spec.path)

        return web.Response(
            text=spec.ui.get_index_page(spec_url=spec_full_path),
            content_type='text/html',
        )

    async def _rpc_handle(self, http_request: web.Request, dispatcher: AioHttpDispatcher) -> web.Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.content_type not in pjrpc.common.REQUEST_CONTENT_TYPES:
            raise web.HTTPUnsupportedMediaType()

        try:
            request_text = await http_request.text()
        except UnicodeDecodeError as e:
            raise web.HTTPBadRequest() from e

        response = await dispatcher.dispatch(request_text, context=http_request)
        if response is None:
            return web.Response()
        else:
            response_text, error_codes = response
            return web.json_response(status=self._status_by_error(error_codes), text=response_text)



from __future__ import annotations

from typing import Any, Awaitable, Callable, Optional, Type, TypeVar, Union

import pjrpc.common.exceptions
from pjrpc.common import Request, Response, UnsetType

__all__ = [
    'AsyncErrorHandlerType',
    'AsyncMiddlewareType',
    'AsyncHandlerType',
    'MiddlewareResponse',
    'MiddlewareType',
    'ErrorHandlerType',
    'ExcludeFunc',
    'ResponseOrUnset',
    'ContextType',
]


ContextType = TypeVar('ContextType')
'''Context argument for RPC methods and middlewares'''  # for sphinx autodoc

ResponseOrUnset = Union[UnsetType, Response]
'''Return value of RPC handlers and middlewares'''  # for sphinx autodoc

AsyncHandlerType = Callable[[Request, ContextType], Awaitable[ResponseOrUnset]]
'''Async RPC handler method, passed to middlewares'''  # for sphinx autodoc

HandlerType = Callable[[Request, ContextType], ResponseOrUnset]
'''Blocking RPC handler method, passed to middlewares'''  # for sphinx autodoc

MiddlewareResponse = Union[UnsetType, Response]
'''middlewares and handlers return Response or UnsetType'''  # for sphinx autodoc

AsyncMiddlewareType = Callable[
    [Request, ContextType, AsyncHandlerType[ContextType]],
    Awaitable[MiddlewareResponse],
]
'''Asynchronous middleware type'''  # for sphinx autodoc

AsyncErrorHandlerType = Callable[
    [Request, ContextType, pjrpc.exceptions.JsonRpcError],
    Awaitable[pjrpc.exceptions.JsonRpcError],
]
'''Asynchronous server error handler'''  # for sphinx autodoc


MiddlewareType = Callable[
    [Request, ContextType, HandlerType[ContextType]],
    MiddlewareResponse,
]
'''Synchronous middleware type'''  # for sphinx autodoc

ErrorHandlerType = Callable[
    [Request, ContextType, pjrpc.exceptions.JsonRpcError],
    pjrpc.exceptions.JsonRpcError,
]
'''Synchronous server error handler'''  # for sphinx autodoc


ExcludeFunc = Callable[[str, Optional[Type[Any]], Optional[Any]], bool]
'''Parameter exclude function'''  # for sphinx autodoc



"""
JSON-RPC version 2.0 protocol implementation.
"""
from __future__ import annotations

import abc
import itertools as it
import operator as op
from typing import Any, ClassVar, Dict, Iterable, Iterator, List, Optional, Set, Tuple, Type

from pjrpc.common.typedefs import Json, JsonRpcParams, JsonRpcRequestId

from .common import UNSET, MaybeSet, UnsetType
from .exceptions import DeserializationError, IdentityError, JsonRpcError


class AbstractRequest(abc.ABC):
    """
    JSON-RPC version 2.0 abstract request.
    """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json_data: Json) -> 'AbstractRequest':
        pass

    @abc.abstractmethod
    def __str__(self) -> str:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    @abc.abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abc.abstractmethod
    def to_json(self) -> Json:
        pass

    @property
    @abc.abstractmethod
    def is_notification(self) -> bool:
        pass


class AbstractResponse(abc.ABC):
    """
    JSON-RPC version 2.0 abstract response.
    """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json_data: Json, error_cls: Type[JsonRpcError] = JsonRpcError) -> 'AbstractResponse':
        pass

    @abc.abstractmethod
    def __str__(self) -> str:
        pass

    @abc.abstractmethod
    def __repr__(self) -> str:
        pass

    @abc.abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @property
    @abc.abstractmethod
    def result(self) -> Any:
        pass

    @property
    @abc.abstractmethod
    def error(self) -> MaybeSet[JsonRpcError]:
        pass

    @property
    @abc.abstractmethod
    def is_success(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def is_error(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def related(self) -> Optional[AbstractRequest]:
        pass

    @related.setter
    def related(self, request: Optional[AbstractRequest]) -> None:
        pass

    @abc.abstractmethod
    def get_error(self) -> JsonRpcError:
        pass

    @abc.abstractmethod
    def to_json(self) -> Json:
        pass


class Response(AbstractResponse):
    """
    JSON-RPC version 2.0 response.

    :param id: response identifier
    :param result: response result
    :param error: response error
    """

    version: ClassVar[str] = '2.0'

    @classmethod
    def from_json(cls, json_data: Json, error_cls: Type[JsonRpcError] = JsonRpcError) -> 'Response':
        """
        Deserializes a response from json data.

        :param json_data: data the response to be deserialized from
        :param error_cls: error class
        :returns: response object
        :raises: :py:class:`pjrpc.common.exceptions.DeserializationError` if format is incorrect
        """

        try:
            if not isinstance(json_data, dict):
                raise DeserializationError("data must be of type dict")

            jsonrpc = json_data['jsonrpc']
            if jsonrpc != cls.version:
                raise DeserializationError(f"jsonrpc version '{json_data['jsonrpc']}' is not supported")

            id = json_data.get('id')
            if id is not None and not isinstance(id, (int, str)):
                raise DeserializationError("field 'id' must be of type integer or string")

            error = json_data.get('error', UNSET)
            if error is not UNSET:
                error = error_cls.from_json(error)

            result = json_data.get('result', UNSET)
            if result is UNSET and error is UNSET:
                raise DeserializationError("'result' or 'error' fields must be provided")
            if result and error:
                raise DeserializationError("'result' and 'error' fields are mutually exclusive")

            return cls(id=id, result=result, error=error)
        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

    def __init__(
        self,
        id: Optional[JsonRpcRequestId],
        result: MaybeSet[Any] = UNSET,
        error: MaybeSet[JsonRpcError] = UNSET,
    ):
        assert result is not UNSET or error is not UNSET, "result or error argument must be provided"
        assert result is UNSET or error is UNSET, "result and error arguments are mutually exclusive"

        self._id = id
        self._result = result
        self._error = error
        self._related: Optional['Request'] = None

    def __str__(self) -> str:
        return f"{self.result if self.is_success else self.error}"

    def __repr__(self) -> str:
        return "{class_name}(id={id}, result={result}, error={error})".format(
            class_name=self.__class__.__name__, id=self._id, result=repr(self._result), error=repr(self._error),
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Response):
            return NotImplemented

        if self.is_success:
            return (self.id, self.result) == (other.id, other.result)
        else:
            return (self.id, self.error) == (other.id, other.error)

    @property
    def id(self) -> Optional[JsonRpcRequestId]:
        """
        Response identifier.
        """

        return self._id

    @property
    def result(self) -> Any:
        """
        Response result. If the response has not succeeded raises an exception deserialized from the `error` field.
        """

        if self._error is not UNSET:
            raise self.get_error()

        return self._result

    @property
    def error(self) -> MaybeSet[JsonRpcError]:
        """
        Response error. If the response has succeeded returns :py:data:`pjrpc.common.common.UNSET`.
        """

        return self._error

    @property
    def is_success(self) -> bool:
        """
        Returns ``True`` if the response has succeeded.
        """

        return self._error is UNSET

    @property
    def is_error(self) -> bool:
        """
        Returns ``True`` if the response has not succeeded.
        """

        return not self.is_success

    @property  # type: ignore[override]
    def related(self) -> Optional['Request']:
        """
        Returns the request related response object if the response has been
        received from the server otherwise returns ``None``.
        """

        return self._related

    @related.setter
    def related(self, request: Optional['Request']) -> None:
        """
        Sets a related request object.
        """

        self._related = request

    def get_error(self) -> JsonRpcError:
        """
        Returns error. If error is not set raises and exception.
        """

        assert not isinstance(self._error, UnsetType), "error is not set"
        return self._error

    def to_json(self) -> Json:
        """
        Serializes the response to json data.

        :returns: json data
        """

        json_data: Dict[str, Any] = {
            'jsonrpc': self.version,
            'id': self._id,
        }
        if self._result is not UNSET:
            json_data.update(result=self.result)
        if self._error is not UNSET:
            json_data.update(error=self.get_error().to_json())

        return json_data


class Request(AbstractRequest):
    """
    JSON-RPC version 2.0 request.

    :param method: method name
    :param params: method parameters
    :param id: request identifier
    """

    version: ClassVar[str] = '2.0'

    @classmethod
    def from_json(cls, json_data: Json) -> 'Request':
        """
        Deserializes a request from json data.

        :param json_data: data the request to be deserialized from
        :returns: request object
        :raises: :py:class:`pjrpc.common.exceptions.DeserializationError` if format is incorrect
        """

        try:
            if not isinstance(json_data, dict):
                raise DeserializationError("data must be of type dict")

            jsonrpc = json_data['jsonrpc']
            if jsonrpc != cls.version:
                raise DeserializationError(f"jsonrpc version '{json_data['jsonrpc']}' is not supported")

            id = json_data.get('id')
            if id is not None and not isinstance(id, (int, str)):
                raise DeserializationError("field 'id' must be of type integer or string")

            method = json_data['method']
            if not isinstance(method, str):
                raise DeserializationError("field 'method' must be of type string")

            params = json_data.get('params', [])
            if not isinstance(params, (list, dict)):
                raise DeserializationError("field 'params' must be of type list or dict")

            return cls(method, params, id)
        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

    @property
    def id(self) -> Optional[JsonRpcRequestId]:
        """
        Request identifier.
        """

        return self._id

    @property
    def method(self) -> str:
        """
        Request method name.
        """

        return self._method

    @property
    def params(self) -> Optional[JsonRpcParams]:
        """
        Request method parameters.
        """

        return self._params

    def __init__(
        self,
        method: str,
        params: Optional[JsonRpcParams] = None,
        id: Optional[JsonRpcRequestId] = None,
    ):
        self._method = method
        self._params = params
        self._id = id

    def __str__(self) -> str:
        if isinstance(self.params, list):
            params = ', '.join(map(str, self.params))
        elif isinstance(self.params, dict):
            params = ','.join(f"{k}={v}" for k, v in self.params.items())
        else:
            params = ''

        return f"{self.method}({params})"

    def __repr__(self) -> str:
        return "{class_name}(method={method}, params={params}, id={id})".format(
            class_name=self.__class__.__name__, method=repr(self._method), params=repr(self._params), id=repr(self._id),
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Request):
            return NotImplemented

        return (self.method, self.params, self.id) == (other.method, other.params, other.id)

    def to_json(self) -> Json:
        """
        Serializes the request to json data.

        :returns: json data
        """

        json_data: Dict[str, Any] = {
            'jsonrpc': self.version,
            'method': self._method,
        }
        if self._id is not None:
            json_data.update(id=self._id)
        if self.params:
            json_data.update(params=self._params)

        return json_data

    @property
    def is_notification(self) -> bool:
        """
        Returns ``True`` if the request is a notification e.g. `id` is ``None``.
        """

        return self.id is None


class BatchResponse(AbstractResponse):
    """
    JSON-RPC 2.0 batch response.

    :param responses: responses to be added to the batch
    :param strict: if ``True`` checks response identifier uniqueness
    """

    version = '2.0'

    @classmethod
    def from_json(cls, json_data: Json, error_cls: Type[JsonRpcError] = JsonRpcError) -> 'BatchResponse':
        """
        Deserializes a batch response from json data.

        :param json_data: data the response to be deserialized from
        :param error_cls: error class
        :returns: batch response object
        """

        try:
            if isinstance(json_data, dict):
                jsonrpc = json_data['jsonrpc']
                if jsonrpc != cls.version:
                    raise DeserializationError(f"jsonrpc version '{json_data['jsonrpc']}' is not supported")

                id, error = json_data.get('id'), json_data.get('error', UNSET)
                if id is None and error is not UNSET:
                    return cls(error=error_cls.from_json(json_data['error']))

            if not isinstance(json_data, (list, tuple)):
                raise DeserializationError("data must be of type list")

        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

        return cls(*(Response.from_json(item) for item in json_data))

    def __init__(self, *responses: Response, error: MaybeSet[JsonRpcError] = UNSET, strict: bool = True):
        self._responses: List[Response] = []
        self._ids: Set[JsonRpcRequestId] = set()
        self._error = error
        self._strict = strict
        self._related: Optional['BatchRequest'] = None

        self.extend(responses)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({','.join(map(repr, self._responses))}, error={repr(self._error)})"

    def __str__(self) -> str:
        return f"[{', '.join(map(str, self._responses))}]" if self.is_success else str(self.error)

    def __getitem__(self, idx: int) -> Response:
        """
        Returns a request at index `idx`.
        """

        return self._responses[idx]

    def __iter__(self) -> Iterator[Response]:
        return iter(self._responses)

    def __len__(self) -> int:
        return len(self._responses)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BatchResponse):
            return NotImplemented

        return all(
            it.starmap(
                op.eq, zip(
                    sorted(self, key=op.attrgetter('id')),
                    sorted(other, key=op.attrgetter('id')),
                ),
            ),
        )

    @property
    def error(self) -> MaybeSet[JsonRpcError]:
        """
        Response error. If the response has succeeded returns :py:data:`pjrpc.common.common.UNSET`.
        """

        return self._error

    @property
    def is_success(self) -> bool:
        """
        Returns ``True`` if the response has succeeded.
        """

        return self._error is UNSET

    @property
    def is_error(self) -> bool:
        """
        Returns ``True`` if the request has not succeeded. Note that it is not the same as
        :py:attr:`pjrpc.common.BatchResponse.has_error`. `is_error` indicates that the batch request failed
        at all, while `has_error` indicates that one of the requests in the batch failed.
        """

        return not self.is_success

    @property
    def has_error(self) -> bool:
        """
        Returns ``True`` if any response has an error.
        """

        return any((response.is_error for response in self._responses))

    @property
    def result(self) -> Tuple[Any, ...]:
        """
        Returns the batch result as a tuple. If any response of the batch has an error
        raises an exception of the first errored response.
        """

        if self.is_error:
            raise self.get_error()

        result = []

        for response in self._responses:
            if response.is_error:
                raise response.get_error()
            result.append(response.result)

        return tuple(result)

    @property  # type: ignore[override]
    def related(self) -> Optional['BatchRequest']:
        """
        Returns the request related response object if the response has been
        received from the server otherwise returns ``None``.
        """

        return self._related

    @related.setter
    def related(self, request: Optional['BatchRequest']) -> None:
        """
        Sets related request object.
        """

        self._related = request

    def get_error(self) -> JsonRpcError:
        """
        Returns error. If error is not set raises and exception.
        """

        assert not isinstance(self._error, UnsetType), "error is not set"
        return self._error

    def append(self, response: Response) -> None:
        """
        Appends a response to the batch.
        """

        self._add_ids(response.id)
        self._responses.append(response)

    def extend(self, responses: Iterable[Response]) -> None:
        """
        Extends the batch with the `responses`.
        """

        self._add_ids(*(resp.id for resp in responses))
        self._responses.extend(responses)

    def to_json(self) -> Json:
        """
        Serializes the batch response to json data.

        :returns: json data
        """

        if self.is_error:
            return Response(id=None, error=self.error).to_json()

        return [response.to_json() for response in self._responses]

    def _add_ids(self, *ids: Optional[JsonRpcRequestId]) -> None:
        if self._strict:
            new_ids = self._ids.copy()

            for id in ids:
                if id is None:
                    continue
                if id in new_ids:
                    raise IdentityError(f"response id duplicates found: {id}")
                else:
                    new_ids.add(id)

            self._ids = new_ids


class BatchRequest(AbstractRequest):
    """
    JSON-RPC 2.0 batch request.

    :param requests: requests to be added to the batch
    :param strict: if ``True`` checks response identifier uniqueness
    """

    version: ClassVar[str] = '2.0'

    @classmethod
    def from_json(cls, data: Json) -> 'BatchRequest':
        """
        Deserializes a batch request from json data.

        :param data: data the request to be deserialized from
        :returns: batch request object
        """

        if not isinstance(data, (list, tuple)):
            raise DeserializationError("data must be of type list")

        if len(data) == 0:
            raise DeserializationError("request list is empty")

        return cls(*(Request.from_json(request) for request in data))

    def __init__(self, *requests: Request, strict: bool = True):
        self._strict = strict
        self._requests: List[Request] = []
        self._ids: Set[JsonRpcRequestId] = set()

        self.extend(requests)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({','.join(map(repr, self._requests))})"

    def __str__(self) -> str:
        return f"[{', '.join(map(str, self._requests))}]"

    def __getitem__(self, idx: int) -> Request:
        """
        Returns a request at index `idx`.
        """

        return self._requests[idx]

    def __iter__(self) -> Iterator[Request]:
        return iter(self._requests)

    def __len__(self) -> int:
        return len(self._requests)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BatchRequest):
            return NotImplemented

        return all(
            it.starmap(
                op.eq, zip(
                    sorted(self, key=op.attrgetter('id')),
                    sorted(other, key=op.attrgetter('id')),
                ),
            ),
        )

    def append(self, request: Request) -> None:
        """
        Appends a request to the batch.
        """

        self._add_ids(request.id)
        self._requests.append(request)

    def extend(self, requests: Iterable[Request]) -> None:
        """
        Extends a batch with `requests`.
        """

        self._add_ids(*(resp.id for resp in requests))
        self._requests.extend(requests)

    def to_json(self) -> Json:
        """
        Serializes the request to json data.

        :returns: json data
        """

        return [request.to_json() for request in self._requests]

    @property
    def is_notification(self) -> bool:
        """
        Returns ``True`` if all the request in the batch are notifications.
        """

        return all(map(op.attrgetter('is_notification'), self._requests))

    def _add_ids(self, *ids: Optional[JsonRpcRequestId]) -> None:
        if self._strict:
            new_ids = self._ids.copy()

            for id in ids:
                if id is None:
                    continue
                if id in new_ids:
                    raise IdentityError(f"request id duplicates found: {id}")
                else:
                    new_ids.add(id)

            self._ids = new_ids



import abc
import enum
import json
from typing import Any, Dict, Iterable, Mapping, Optional

from pjrpc.server import Method


class JSONEncoder(json.JSONEncoder):
    """
    Schema JSON encoder.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, enum.Enum):
            return o.value

        return super().default(o)


class BaseUI(abc.ABC):
    """
    Base UI.
    """

    @abc.abstractmethod
    def get_static_folder(self) -> str:
        """
        Returns ui statics folder.
        """

    @abc.abstractmethod
    def get_index_page(self, spec_url: str) -> str:
        """
        Returns ui index webpage.

        :param spec_url: specification url.
        """


class Specification(abc.ABC):
    """
    JSON-RPC specification.

    :param path: specification url path suffix
    :param ui: specification ui instance
    :param ui_path: specification ui url path suffix
    """

    def __init__(self, path: str = '/spec.json', ui: Optional[BaseUI] = None, ui_path: Optional[str] = None):
        self._path = path
        self._ui = ui
        self._ui_path = ui_path

    @property
    def path(self) -> str:
        """
        Returns specification url path.
        """

        return self._path

    @property
    def ui(self) -> Optional[BaseUI]:
        """
        Returns ui instance.
        """

        return self._ui

    @property
    def ui_path(self) -> Optional[str]:
        """
        Returns specification ui url path.
        """

        return self._ui_path

    @abc.abstractmethod
    def schema(
        self,
        path: str,
        methods_map: Mapping[str, Iterable[Method]] = {},
    ) -> Dict[str, Any]:
        """
        Returns specification schema.

        :param path: methods endpoint path
        :param methods_map: methods map the specification is generated for.
                            Each item is a mapping from a prefix to methods on which the methods will be served
        """



from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

import docstring_parser

from pjrpc.common import UNSET, MaybeSet, exceptions
from pjrpc.common.typedefs import MethodType
from pjrpc.server.specs.extractors import BaseSchemaExtractor, JsonRpcError
from pjrpc.server.specs.schemas import build_request_schema, build_response_schema
from pjrpc.server.typedefs import ExcludeFunc


class DocstringSchemaExtractor(BaseSchemaExtractor):
    """
    docstring method specification generator.

    :param exclude_param: a function that decides if the parameters must be excluded
                          from schema (useful for dependency injection)
    """

    def __init__(self, exclude_param: Optional[ExcludeFunc] = None):
        self._exclude_param = exclude_param or (lambda *args: False)

    def extract_params_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            exclude: Iterable[str] = (),
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        exclude = set(exclude)
        parameters_schema = {}

        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            for param in doc.params:
                if param.arg_name in exclude or self._exclude_param(param.arg_name, None, None):
                    continue

                parameters_schema[param.arg_name] = {
                    'title': param.arg_name.capitalize(),
                    'description': param.description if param.description is not None else UNSET,
                    'type': param.type_name,
                }

        return parameters_schema, {}

    def extract_request_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            exclude: Iterable[str] = (),
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        exclude = set(exclude)
        parameters_schema, components = self.extract_params_schema(method_name, method, ref_template, exclude)

        return build_request_schema(method_name, parameters_schema), components

    def extract_result_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        result_schema = {}

        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            if doc and doc.returns:
                result_schema = {
                    'type': doc.returns.type_name,
                    'title': 'Result',
                    'description': doc.returns.description if doc.returns.description is not None else UNSET,
                }

        return result_schema, {}

    def extract_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        result_schema, components = self.extract_result_schema(method_name, method, ref_template)

        return build_response_schema(result_schema, errors or []), components

    def extract_errors(self, method: MethodType) -> MaybeSet[List[Type[JsonRpcError]]]:
        errors_schema: List[Type[JsonRpcError]] = []

        if method.__doc__:
            error_map = {
                error.__name__: error
                for error in exceptions.JsonRpcErrorMeta.__errors_mapping__.values()
            }

            doc = docstring_parser.parse(method.__doc__)
            for error in doc.raises:
                error_cls = error_map.get(error.type_name or '')
                if error_cls:
                    errors_schema.append(error_cls)

        return errors_schema or UNSET

    def extract_description(self, method: MethodType) -> MaybeSet[str]:
        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            description = doc.long_description or UNSET
        else:
            description = UNSET

        return description

    def extract_summary(self, method: MethodType) -> MaybeSet[str]:
        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            description = doc.short_description or UNSET
        else:
            description = UNSET

        return description

    def extract_deprecation_status(self, method: MethodType) -> MaybeSet[bool]:
        is_deprecated: MaybeSet[bool]
        if method.__doc__:
            doc = docstring_parser.parse(method.__doc__)
            is_deprecated = bool(doc.deprecation)
        else:
            is_deprecated = UNSET

        return is_deprecated



import inspect
from typing import Any, Dict, Generic, Iterable, Literal, Optional, Tuple, Type, TypeVar, Union

import pydantic as pd

from pjrpc.common import exceptions
from pjrpc.common.typedefs import MethodType
from pjrpc.server.specs.extractors import BaseSchemaExtractor
from pjrpc.server.typedefs import ExcludeFunc


def to_camel(string: str) -> str:
    return ''.join(word.capitalize() for word in string.split('_'))


MethodT = TypeVar('MethodT', bound=str)
ParamsT = TypeVar('ParamsT', bound=pd.BaseModel, covariant=True)


class JsonRpcRequest(
    pd.BaseModel, Generic[MethodT, ParamsT],
    title="Request",
    extra='forbid', strict=True,
):
    jsonrpc: Literal['2.0', '1.0'] = pd.Field(title="Version", description="JSON-RPC protocol version")
    id: Optional[Union[str, int]] = pd.Field(None, description="Request identifier", examples=[1])
    method: MethodT = pd.Field(description="Method name")
    params: ParamsT = pd.Field(description="Method parameters")


RequestT = TypeVar('RequestT', bound=Union[JsonRpcRequest[Any, Any]])


class JsonRpcRequestWrapper(pd.RootModel[RequestT], Generic[RequestT], title="Request"):
    pass


ResultT = TypeVar('ResultT', covariant=True)


class JsonRpcResponseSuccess(
    pd.BaseModel, Generic[ResultT],
    title='Success',
    extra='forbid', strict=True,
):
    jsonrpc: Literal['2.0', '1.0'] = pd.Field(title="Version", description="JSON-RPC protocol version")
    id: Union[str, int] = pd.Field(description="Request identifier", examples=[1])
    result: ResultT = pd.Field(description="Method execution result")


ErrorCodeT = TypeVar('ErrorCodeT', bound=int)
ErrorDataT = TypeVar('ErrorDataT')


class JsonRpcError(
    pd.BaseModel, Generic[ErrorCodeT, ErrorDataT],
    title="Error",
    extra='forbid', strict=True,
):
    code: ErrorCodeT = pd.Field(description="Error code")
    message: str = pd.Field(description="Error message")
    data: ErrorDataT = pd.Field(description="Error additional data")


JsonRpcErrorT = TypeVar('JsonRpcErrorT', bound=JsonRpcError[Any, Any], covariant=True)


class JsonRpcResponseError(
    pd.BaseModel, Generic[JsonRpcErrorT],
    title="ResponseError",
    extra='forbid', strict=True,
):
    jsonrpc: Literal['1.0', '2.0'] = pd.Field(title="Version", description="JSON-RPC protocol version")
    id: Union[str, int] = pd.Field(description="Request identifier", examples=[1])
    error: JsonRpcErrorT = pd.Field(description="Request error")


ResponseT = TypeVar('ResponseT', bound=Union[JsonRpcResponseSuccess[Any], JsonRpcResponseError[Any]])


class JsonRpcResponseWrapper(pd.RootModel[ResponseT], Generic[ResponseT], title="Response"):
    pass


class PydanticSchemaExtractor(BaseSchemaExtractor):
    """
    Pydantic method specification extractor.

    :param config_args: model configuration parameters
    :param exclude_param: a function that decides if the parameters must be excluded
                          from schema (useful for dependency injection)
    """

    def __init__(self, exclude_param: Optional[ExcludeFunc] = None, **config_args: Any):
        self._exclude_param = exclude_param or (lambda *args: False)
        self._config_args = config_args

    def extract_params_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            exclude: Iterable[str] = (),
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        params_model = self._build_params_model(method_name, method, exclude)
        params_schema = params_model.model_json_schema(ref_template=ref_template)

        return params_schema, params_schema.pop('$defs', {})

    def extract_request_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            exclude: Iterable[str] = (),
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        exclude = set(exclude)

        params_model = self._build_params_model(method_name, method, exclude)
        request_model = pd.create_model(
            f'{to_camel(method_name)}Request',
            __base__=JsonRpcRequestWrapper[
                JsonRpcRequest[Literal[method_name], params_model],  # type: ignore[valid-type]
            ],
            __cls_kwargs__=self._config_args,
        )
        request_schema = request_model.model_json_schema(ref_template=ref_template)

        return request_schema, request_schema.pop('$defs', {})

    def extract_result_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        result_model = self._build_result_model(method_name, method)
        result_schema = result_model.model_json_schema(ref_template=ref_template)

        return result_schema, result_schema.pop('$defs', {})

    def extract_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[Type[exceptions.JsonRpcError]]] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        return_model = self._build_result_model(method_name, method)

        response_model: Type[pd.BaseModel]
        error_models = tuple(
            pd.create_model(
                error.__name__,
                __base__=JsonRpcResponseError[JsonRpcError[Literal[error.code], Any]],  # type: ignore[name-defined]
                __cls_kwargs__=dict(
                    self._config_args,
                    title=error.__name__,
                    json_schema_extra=dict(description=f'**{error.code}** {error.message}'),
                ),
            ) for error in errors or []
        )

        response_model = pd.create_model(
            f'{to_camel(method_name)}Response',
            __base__=JsonRpcResponseWrapper[
                Union[(JsonRpcResponseSuccess[return_model], *error_models)],  # type: ignore[valid-type]
            ],
        )
        response_schema = response_model.model_json_schema(ref_template=ref_template)
        if error_models:
            response_schema['description'] = '\n'.join(
                f'* {error.model_json_schema().get("description", error.__name__)}'
                for error in error_models
            )

        return response_schema, response_schema.pop('$defs', {})

    def extract_error_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[Type[exceptions.JsonRpcError]]] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        error_models = tuple(
            pd.create_model(
                error.__name__,
                __base__=JsonRpcResponseError[JsonRpcError[Literal[error.code], Any]],  # type: ignore[name-defined]
                __cls_kwargs__=dict(
                    self._config_args,
                    title=error.__name__,
                    json_schema_extra=dict(description=f'**{error.code}** {error.message}'),
                ),
            ) for error in errors or []
        )
        if len(error_models) == 1:
            response_model = pd.create_model(
                f'{to_camel(method_name)}Response',
                __base__=JsonRpcResponseWrapper[Union[error_models]],  # type: ignore[valid-type]
            )
            response_schema = response_model.model_json_schema(ref_template=ref_template)
        else:
            response_model = pd.create_model(
                f'{to_camel(method_name)}Response',
                __base__=JsonRpcResponseWrapper[Union[error_models]],  # type: ignore[valid-type]
            )
            response_schema = response_model.model_json_schema(ref_template=ref_template)

        if error_models:
            response_schema['description'] = '\n'.join(
                f'* {error.model_json_schema().get("description", error.__name__)}'
                for error in error_models
            )
        else:
            response_schema['description'] = 'Error'

        return response_schema, response_schema.pop('$defs', {})

    def _build_params_model(
            self,
            method_name: str,
            method: MethodType,
            exclude: Iterable[str] = (),
    ) -> Type[pd.BaseModel]:
        signature = inspect.signature(method)

        field_definitions: Dict[str, Any] = {}
        for param in signature.parameters.values():
            if param.name in exclude or self._exclude_param(param.name, param.annotation, param.default):
                continue

            if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]:
                field_definitions[param.name] = (
                    param.annotation if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else ...,
                )

        return pd.create_model(
            f'{to_camel(method_name)}Parameters',
            **field_definitions,
            __cls_kwargs__=dict(self._config_args, extra='forbid'),
        )

    def _build_result_model(self, method_name: str, method: MethodType) -> Type[pd.BaseModel]:
        result = inspect.signature(method)

        if result.return_annotation is inspect.Parameter.empty:
            return_annotation = Any
        elif result.return_annotation is None:
            return_annotation = Optional[None]
        else:
            return_annotation = result.return_annotation

        return pd.create_model(
            f'{to_camel(method_name)}Result',
            __base__=pd.RootModel[return_annotation],  # type: ignore[valid-type]
            __cls_kwargs__=dict(self._config_args),
        )



__title__ = 'pjrpc'
__description__ = 'Extensible JSON-RPC library'
__url__ = 'https://github.com/dapper91/pjrpc'

__version__ = '1.13.0'

__author__ = 'Dmitry Pershin'
__email__ = 'dapper1291@gmail.com'

__license__ = 'Public Domain License'



import copy
from typing import Any, Dict, Iterable, List, Type

from pjrpc.common.exceptions import JsonRpcError

REQUEST_SCHEMA: Dict[str, Any] = {
    'title': 'Request',
    'type': 'object',
    'properties': {
        'jsonrpc': {
            'title': 'Version',
            'description': 'JSON-RPC protocol version',
            'type': 'string',
            'enum': ['2.0', '1.0'],
        },
        'id': {
            'title': 'Id',
            'description': 'Request identifier',
            'anyOf': [
                {'type': 'string'},
                {'type': 'integer'},
                {'type': 'null'},
            ],
            'examples': [1],
            'default': None,
        },
        'method': {
            'title': 'Method',
            'description': 'Method name',
            'type': 'string',
        },
        'params': {
            'title': 'Parameters',
            'description': 'Method parameters',
            'type': 'object',
            'properties': {},
        },
    },
    'required': ['jsonrpc', 'method', 'params'],
    'additionalProperties': False,
}

RESULT_SCHEMA: Dict[str, Any] = {
    'title': 'Success',
    'type': 'object',
    'properties': {
        'jsonrpc': {
            'title': 'Version',
            'description': 'JSON-RPC protocol version',
            'type': 'string',
            'enum': ['2.0', '1.0'],
        },
        'id': {
            'title': 'Id',
            'description': 'Request identifier',
            'anyOf': [
                {'type': 'string'},
                {'type': 'integer'},
            ],
            'examples': [1],
        },
        'result': {},
    },
    'required': ['jsonrpc', 'id', 'result'],
    'additionalProperties': False,
}
ERROR_SCHEMA: Dict[str, Any] = {
    'title': 'Error',
    'type': 'object',
    'properties': {
        'jsonrpc': {
            'title': 'Version',
            'description': 'JSON-RPC protocol version',
            'type': 'string',
            'enum': ['2.0', '1.0'],
        },
        'id': {
            'title': 'Id',
            'description': 'Request identifier',
            'anyOf': [
                {'type': 'string'},
                {'type': 'integer'},
            ],
            'examples': [1],
        },
        'error': {
            'type': 'object',
            'properties': {
                'code': {
                    'title': 'Code',
                    'description': 'Error code',
                    'type': 'integer',
                },
                'message': {
                    'title': 'Message',
                    'description': 'Error message',
                    'type': 'string',
                },
                'data': {
                    'title': 'Data',
                    'description': 'Error additional data',
                    'type': 'object',
                },
            },
            'required': ['code', 'message'],
            'additionalProperties': False,
        },
    },
    'required': ['jsonrpc', 'error'],
    'additionalProperties': False,
}


def build_request_schema(method_name: str, parameters_schema: Dict[str, Any]) -> Dict[str, Any]:
    reqeust_schema = copy.deepcopy(REQUEST_SCHEMA)

    reqeust_schema['properties']['method']['const'] = method_name
    reqeust_schema['properties']['params'] = {
        'title': 'Parameters',
        'description': 'Reqeust parameters',
        'type': 'object',
        'properties': parameters_schema,
        'additionalProperties': False,
    }

    return reqeust_schema


def build_response_schema(result_schema: Dict[str, Any], errors: Iterable[Type[JsonRpcError]]) -> Dict[str, Any]:
    response_schema = copy.deepcopy(RESULT_SCHEMA)
    response_schema['properties']['result'] = result_schema

    if errors:
        error_schemas: List[Dict[str, Any]] = []
        for error in errors:
            error_schema = copy.deepcopy(ERROR_SCHEMA)
            error_props = error_schema['properties']['error']['properties']
            error_props['code']['const'] = error.code
            error_props['message']['const'] = error.message
            error_schemas.append(error_schema)

        response_schema = {'oneOf': [response_schema] + error_schemas}

    return response_schema



"""
OpenRPC specification generator. See https://spec.open-rpc.org/.
"""
import copy
import enum

try:
    import dataclasses as dc
except ImportError:
    raise AssertionError("python 3.7 or later is required")

from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Type, TypedDict, Union

from pjrpc.common import UNSET, MaybeSet, UnsetType, exceptions
from pjrpc.common.typedefs import Func
from pjrpc.server import Method, utils

from . import Specification, extractors

Json = Union[str, int, float, dict, bool, list, tuple, set, None]  # type: ignore[type-arg]
JsonSchema = Dict[str, Any]


def remove_prefix(s: str, prefix: str) -> str:
    return s[len(prefix):] if s.startswith(prefix) else s


def follow_ref(schema: Dict[str, Any], components: Dict[str, Dict[str, Any]], ref_prefix: str) -> Dict[str, Any]:
    if '$ref' in schema:
        ref = remove_prefix(schema['$ref'], ref_prefix)
        schema = components[ref]

    return schema


@dc.dataclass
class Reference:
    """
    A simple object to allow referencing other components in the specification, internally and externally.
    :param ref: the reference identifier.
    """

    ref: str

    def __post_init__(self) -> None:
        self.__dict__['$ref'] = self.__dict__['ref']
        self.__dataclass_fields__['ref'].name = '$ref'  # noqa


@dc.dataclass
class Contact:
    """
    Contact information for the exposed API.

    :param name: the identifying name of the contact person/organization
    :param url: the URL pointing to the contact information
    :param email: the email address of the contact person/organization
    """

    name: MaybeSet[str] = UNSET
    url: MaybeSet[str] = UNSET
    email: MaybeSet[str] = UNSET


@dc.dataclass
class License:
    """
    License information for the exposed API.

    :param name: the license name used for the API
    :param url: a URL to the license used for the API
    """

    name: str
    url: MaybeSet[str] = UNSET


@dc.dataclass
class Info:
    """
    Metadata about the API.

    :param title: the title of the application
    :param version: the version of the OpenRPC document
    :param description: a verbose description of the application
    :param contact: the contact information for the exposed API
    :param license: the license information for the exposed API
    :param termsOfService: a URL to the Terms of Service for the API
    """

    title: str
    version: str
    description: MaybeSet[str] = UNSET
    contact: MaybeSet[Contact] = UNSET
    license: MaybeSet[License] = UNSET
    termsOfService: MaybeSet[str] = UNSET


@dc.dataclass
class ServerVariable:
    """
    An object representing a Server Variable for server URL template substitution.

    :param default: The default value to use for substitution.
    :param enum: An enumeration of string values to be used if the substitution options are from a limited set.
    :param description: An optional description for the server variable.
    """

    default: str
    enum: MaybeSet[List[str]] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Server:
    """
    Connectivity information of a target server.

    :param name: a name to be used as the canonical name for the server.
    :param url: a URL to the target host. This URL supports Server Variables.
    :param summary: a short summary of what the server is
    :param description: an optional string describing the host designated by the URL
    :param variables: A map between a variable name and its value.
                      The value is passed into the Runtime Expression to produce a server URL.
    """

    name: str
    url: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    variables: MaybeSet[Dict[str, ServerVariable]] = UNSET


@dc.dataclass
class ExternalDocumentation:
    """
    Allows referencing an external resource for extended documentation.

    :param url: A verbose explanation of the target documentation
    :param description: The URL for the target documentation. Value MUST be in the format of a URL
    """

    url: str
    description: MaybeSet[str] = UNSET


@dc.dataclass
class Tag:
    """
    A list of tags for API documentation control.
    Tags can be used for logical grouping of methods by resources or any other qualifier.

    :param name: the name of the tag
    :param description: a verbose explanation for the tag
    :param externalDocs: additional external documentation for this tag
    """

    name: str
    description: MaybeSet[str] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET


@dc.dataclass
class ExampleObject:
    """
    The ExampleObject object is an object the defines an example.

    :param value: embedded literal example
    :param name: canonical name of the example
    :param summary: short description for the example
    :param description: a verbose explanation of the example
    :param externalValue: a URL that points to the literal example.
    """

    value: Any
    name: str
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    externalValue: MaybeSet[str] = UNSET


@dc.dataclass
class MethodExample:
    """
    The example Pairing object consists of a set of example params and result.

    :param params: example parameters
    :param result: example result
    :param name: name for the example pairing
    :param summary: short description for the example pairing
    :param description: a verbose explanation of the example pairing
    """

    name: str
    params: List[Union[ExampleObject, Reference]]
    result: Union[ExampleObject, Reference]
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET


@dc.dataclass
class ContentDescriptor:
    """
    Content Descriptors are objects that describe content.
    They are reusable ways of describing either parameters or result.

    :param name: name of the content that is being described
    :param schema: schema that describes the content. The Schema Objects MUST follow the specifications outline
                   in the JSON Schema Specification 7 (https://json-schema.org/draft-07/json-schema-release-notes.html)
    :param summary: a short summary of the content that is being described
    :param description: a verbose explanation of the content descriptor behavior
    :param required: determines if the content is a required field
    :param deprecated: specifies that the content is deprecated and SHOULD be transitioned out of usage
    """

    name: str
    schema: JsonSchema
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    required: MaybeSet[bool] = UNSET
    deprecated: MaybeSet[bool] = UNSET


@dc.dataclass
class Error:
    """
    Defines an application level error.

    :param code: a Number that indicates the error type that occurred
    :param message: a String providing a short description of the error
    :param data: a Primitive or Structured value that contains additional information about the error
    """

    code: int
    message: str
    data: MaybeSet[Any] = UNSET


class ParamStructure(str, enum.Enum):
    """
    The expected format of the parameters.
    """

    BY_NAME = 'by-name'
    BY_POSITION = 'by-position'
    EITHER = 'either'


@dc.dataclass
class Link:
    """
    :param name: Canonical name of the link.
    :param description: A description of the link.
    :param summary: Short description for the link.
    :param method: The name of an existing, resolvable OpenRPC method, as defined with a unique method.
    :param params: A map representing parameters to pass to a method as specified with method.
                   The key is the parameter name to be used, whereas the value can be a constant or a runtime
                   expression to be evaluated and passed to the linked method.
    :param server: A server object to be used by the target method.
    """

    name: str
    description: MaybeSet[str] = UNSET
    summary: MaybeSet[str] = UNSET
    method: MaybeSet[str] = UNSET
    params: MaybeSet[Any] = UNSET
    server: MaybeSet[Server] = UNSET


@dc.dataclass
class MethodInfo:
    """
    Describes the interface for the given method name.

    :param name: the canonical name for the method
    :param params: a list of parameters that are applicable for this method
    :param result: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage examples
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    :param paramStructure: the expected format of the parameters
    :param externalDocs: additional external documentation for this method
    :param servers: an alternative servers array to service this method
    """

    name: str
    params: List[Union[ContentDescriptor, Reference]]
    result: Union[ContentDescriptor, Reference]
    errors: MaybeSet[List[Union[Error, Reference]]] = UNSET
    links: MaybeSet[List[Union[Link, Reference]]] = UNSET
    paramStructure: MaybeSet[ParamStructure] = UNSET
    examples: MaybeSet[List[Union[MethodExample, Reference]]] = UNSET
    summary: MaybeSet[str] = UNSET
    description: MaybeSet[str] = UNSET
    tags: MaybeSet[List[Union[Tag, Reference]]] = UNSET
    deprecated: MaybeSet[bool] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    servers: MaybeSet[List[Server]] = UNSET


@dc.dataclass
class Components:
    """
    Set of reusable objects for different aspects of the OpenRPC.

    :param contentDescriptors: reusable Schema Objects
    :param schemas: reusable Schema Objects
    :param examples: reusable Schema Objects
    :param links: reusable Schema Objects
    :param errors: reusable Schema Objects
    :param examplePairingObjects: reusable Schema Objects
    :param tags: reusable Schema Objects
    """

    contentDescriptors: MaybeSet[Dict[str, ContentDescriptor]] = UNSET
    schemas: MaybeSet[Dict[str, JsonSchema]] = UNSET
    examples: MaybeSet[Dict[str, ExampleObject]] = UNSET
    links: MaybeSet[Dict[str, Link]] = UNSET
    errors: MaybeSet[Dict[str, Error]] = UNSET
    examplePairingObjects: MaybeSet[Dict[str, MethodExample]] = UNSET
    tags: MaybeSet[Dict[str, Tag]] = UNSET


@dc.dataclass
class SpecRoot:
    """
    The root object of the OpenRPC document.

    :param info: provides metadata about the API
    :param components: an element to hold various schemas for the specification.
    :param methods: the available methods for the API
    :param servers: connectivity information
    :param externalDocs: additional external documentation
    :param openrpc: the semantic version number of the OpenRPC Specification version that the OpenRPC document uses
    """

    info: Info
    components: Components
    methods: List[Union[MethodInfo, Reference]] = dc.field(default_factory=list)
    servers: MaybeSet[List[Server]] = UNSET
    externalDocs: MaybeSet[ExternalDocumentation] = UNSET
    openrpc: str = '1.3.2'


class OpenRpcMeta(TypedDict):
    params_schema: MaybeSet[List[ContentDescriptor]]
    result_schema: MaybeSet[ContentDescriptor]
    errors: MaybeSet[List[Error]]
    examples: MaybeSet[List[Union[MethodExample, Reference]]]
    tags: MaybeSet[List[Union[Tag, Reference]]]
    summary: MaybeSet[str]
    description: MaybeSet[str]
    deprecated: MaybeSet[bool]
    external_docs: MaybeSet[ExternalDocumentation]
    servers: MaybeSet[List[Server]]


def annotate(
    params_schema: MaybeSet[List[ContentDescriptor]] = UNSET,
    result_schema: MaybeSet[ContentDescriptor] = UNSET,
    errors: MaybeSet[List[Union[Error, Type[exceptions.JsonRpcError]]]] = UNSET,
    examples: MaybeSet[List[Union[MethodExample, Reference]]] = UNSET,
    summary: MaybeSet[str] = UNSET,
    description: MaybeSet[str] = UNSET,
    tags: MaybeSet[List[Union[Tag, str]]] = UNSET,
    deprecated: MaybeSet[bool] = UNSET,
    external_docs: MaybeSet[ExternalDocumentation] = UNSET,
    servers: MaybeSet[List[Server]] = UNSET,
) -> Callable[[Func], Func]:
    """
    Adds JSON-RPC method to the API specification.

    :param params_schema: a list of parameters that are applicable for this method
    :param result_schema: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage example
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    :param external_docs: additional external documentation for the method
    :param servers: an alternative servers array to service this method
    """

    def decorator(method: Func) -> Func:
        meta: OpenRpcMeta = dict(
            params_schema=params_schema,
            result_schema=result_schema,
            errors=[
                error if isinstance(error, Error) else Error(code=error.code, message=error.message)
                for error in errors
            ] if errors else UNSET,
            examples=examples,
            tags=[
                Tag(name=tag) if isinstance(tag, str) else tag
                for tag in tags
            ] if tags else UNSET,
            summary=summary,
            description=description,
            deprecated=deprecated,
            external_docs=external_docs,
            servers=servers,
        )
        utils.set_meta(method, openrpc_spec=meta)

        return method

    return decorator


@dc.dataclass(init=False)
class OpenRPC(Specification):
    """
    OpenRPC Specification.

    :param info: provides metadata about the API
    :param path: specification url path
    :param servers: connectivity information
    :param external_docs: additional external documentation
    :param openrpc: the semantic version number of the OpenRPC Specification version that the OpenRPC document uses
    :param schema_extractor: method specification extractor
    """

    def __init__(
        self,
        info: Info,
        path: str = '/openrpc.json',
        servers: MaybeSet[List[Server]] = UNSET,
        external_docs: MaybeSet[ExternalDocumentation] = UNSET,
        openrpc: str = '1.3.2',
        schema_extractor: Optional[extractors.BaseSchemaExtractor] = None,
    ):
        super().__init__(path)

        self._spec = SpecRoot(
            info=info,
            servers=servers,
            externalDocs=external_docs,
            openrpc=openrpc,
            methods=[],
            components=Components(),
        )
        self._schema_extractor = schema_extractor or extractors.BaseSchemaExtractor()

    def schema(
        self,
        path: str,
        methods_map: Mapping[str, Iterable[Method]] = {},
    ) -> Dict[str, Any]:
        spec = copy.deepcopy(self._spec)

        spec.components.schemas = {}

        for method in methods_map.get('', []):
            summary, description = self._extract_description(method)
            params_schema = self._extract_params_schema(spec, method)
            result_schema = self._extract_result_schema(spec, method)
            errors = self._extract_errors(method)
            deprecated = self._extract_deprecated(method)
            tags = self._extract_tags(method)
            external_docs = self._extract_external_docs(method)
            servers = self._extract_servers(method)
            examples = self._extract_examples(method)

            spec.methods.append(
                MethodInfo(
                    name=method.name,
                    params=list(params_schema),
                    result=result_schema,
                    errors=list(errors) if errors else UNSET,
                    examples=examples or UNSET,
                    summary=summary,
                    description=description,
                    tags=tags or UNSET,
                    deprecated=deprecated,
                    servers=servers or UNSET,
                    externalDocs=external_docs or UNSET,
                ),
            )

        return dc.asdict(
            spec,
            dict_factory=lambda iterable: dict(
                filter(lambda item: not isinstance(item[1], UnsetType), iterable),
            ),
        )

    def _extract_params_schema(self, spec: SpecRoot, method: Method) -> List[ContentDescriptor]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        if not (params_descriptors := annotations.get('params_schema')):
            request_ref_prefix = '#/components/schemas/'
            params_schema, components = self._schema_extractor.extract_params_schema(
                method.name,
                method.method,
                ref_template=f'{request_ref_prefix}{{model}}',
                exclude=[method.context] if method.context else [],
            )
            params_descriptors = [
                ContentDescriptor(
                    name=name,
                    schema=schema,
                    summary=schema.get('title', UNSET),
                    description=schema.get('description', UNSET),
                    required=name in params_schema.get('required', []),
                    deprecated=schema.get('deprecated', UNSET),
                ) for name, schema in params_schema['properties'].items()
            ]

            spec.components.schemas = schemas = spec.components.schemas or {}
            schemas.update(components)

        return params_descriptors

    def _extract_result_schema(self, spec: SpecRoot, method: Method) -> ContentDescriptor:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        if not (result_descriptor := annotations.get('result_schema')):
            response_ref_prefix = '#/components/schemas/'
            result_schema, components = self._schema_extractor.extract_result_schema(
                method.name,
                method.method,
                ref_template=f'{response_ref_prefix}{{model}}',
            )
            result_descriptor = ContentDescriptor(
                name='result',
                schema=result_schema,
                summary=result_schema.get('title', UNSET),
                description=result_schema.get('description', UNSET),
                required='result' in result_schema.get('required', []),
                deprecated=result_schema.get('deprecated', UNSET),
            )
            spec.components.schemas = schemas = spec.components.schemas or {}
            schemas.update(components)

        return result_descriptor

    def _extract_errors(self, method: Method) -> MaybeSet[List[Error]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        errors = annotations.get('errors', UNSET) or []
        errors.extend([
            Error(code=error.code, message=error.message)
            for error in self._schema_extractor.extract_errors(method.method) or []
        ])

        unique_errors = list({error.code: error for error in errors}.values())

        return unique_errors or UNSET

    def _extract_description(self, method: Method) -> Tuple[MaybeSet[str], MaybeSet[str]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        summary = annotations.get('summary', UNSET) or self._schema_extractor.extract_summary(method.method)
        description = annotations.get('description', UNSET) or self._schema_extractor.extract_description(method.method)

        return summary, description

    def _extract_tags(self, method: Method) -> MaybeSet[List[Union[Tag, Reference]]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        tags = annotations.get('tags', UNSET) or []

        return tags or UNSET

    def _extract_servers(self, method: Method) -> MaybeSet[List[Server]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        servers = annotations.get('servers', UNSET) or []

        return servers or UNSET

    def _extract_deprecated(self, method: Method) -> MaybeSet[bool]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        deprecated = annotations.get('deprecated', UNSET)
        if deprecated is UNSET:
            deprecated = self._schema_extractor.extract_deprecation_status(method.method)

        return deprecated

    def _extract_external_docs(self, method: Method) -> MaybeSet[ExternalDocumentation]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        external_docs = annotations.get('external_docs', UNSET)

        return external_docs

    def _extract_examples(self, method: Method) -> MaybeSet[List[Union[MethodExample, Reference]]]:
        method_meta = utils.get_meta(method.method)
        annotations: OpenRpcMeta = method_meta.get('openrpc_spec', {})

        examples = annotations.get('examples', UNSET)

        return examples



import inspect
import itertools as it
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

from pjrpc.common import UNSET, MaybeSet, UnsetType
from pjrpc.common.exceptions import JsonRpcError
from pjrpc.common.typedefs import MethodType


class BaseSchemaExtractor:
    """
    Base method schema extractor.
    """

    def extract_params_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            exclude: Iterable[str] = (),
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        Extracts params schema.
        """

        return {}, {}

    def extract_request_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            exclude: Iterable[str] = (),
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        Extracts request schema.
        """

        return {}, {}

    def extract_result_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        Extracts result schema.
        """

        return {}, {}

    def extract_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        Extracts response schema.
        """

        return {}, {}

    def extract_error_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[Type[JsonRpcError]]] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        Extracts error response schema.
        """

        return {}, {}

    def extract_description(self, method: MethodType) -> MaybeSet[str]:
        """
        Extracts method description.
        """

        description: MaybeSet[str]
        if method.__doc__:
            doc = inspect.cleandoc(method.__doc__)
            description = '\n'.join(it.takewhile(lambda line: line, doc.split('\n')))
        else:
            description = UNSET

        return description

    def extract_summary(self, method: MethodType) -> MaybeSet[str]:
        """
        Extracts method summary.
        """

        description = self.extract_description(method)

        summary: MaybeSet[str]
        if not isinstance(description, UnsetType):
            summary = description.split('.')[0]
        else:
            summary = UNSET

        return summary

    def extract_errors(self, method: MethodType) -> MaybeSet[List[Type[JsonRpcError]]]:
        """
        Extracts method errors.
        """

        return UNSET

    def extract_deprecation_status(self, method: MethodType) -> MaybeSet[bool]:
        """
        Extracts method deprecation status.
        """

        return UNSET



import functools as ft
import json
from typing import Any, Callable, List, Optional, cast

import django.utils.functional
import django.utils.log
from django import urls
from django.conf import settings
from django.conf.urls.static import static
from django.http import HttpRequest, HttpResponse, HttpResponseBadRequest, HttpResponseNotAllowed
from django.utils.module_loading import import_string
from django.views.decorators.csrf import csrf_exempt

import pjrpc.server
from pjrpc.common.typedefs import Func
from pjrpc.server import specs, utils

DjangoDispatcher = pjrpc.server.Dispatcher[HttpRequest]


def require_http_methods(request_method_list: List[str]) -> Callable[[Func], Func]:
    def decorator(func: Func) -> Func:
        @ft.wraps(func)
        def inner(self: 'JsonRPCSite', request: HttpRequest, *args: Any, **kwargs: Any) -> Any:
            if request.method not in request_method_list:
                response = HttpResponseNotAllowed(request_method_list)
                django.utils.log.log_response(
                    'Method Not Allowed (%s): %s', request.method, request.path,
                    response=response,
                    request=request,
                )
                return response
            return func(self, request, *args, **kwargs)
        return cast(Func, inner)
    return decorator


class JsonRPCSite:

    def __init__(self, path: str = '',  spec: Optional[specs.Specification] = None, **kwargs: Any):
        self._path = path
        self._spec = spec

        self._dispatcher = DjangoDispatcher(**kwargs)
        self._urls = [
            urls.path(path, self._rpc_handle),
        ]

        if self._spec:
            self._urls.append(urls.path(utils.join_path(self._path, self._spec.path), self._generate_spec))

            if self._spec.ui and self._spec.ui_path:
                path = utils.join_path(self._path, self._spec.ui_path)
                self._urls.extend((
                    urls.path(utils.join_path(path, '/'), self._ui_index_page),
                    urls.path(utils.join_path(path, 'index.html'), self._ui_index_page),
                ))
                self._urls.extend(static(path, document_root=str(self._spec.ui.get_static_folder())))

    @property
    def dispatcher(self) -> DjangoDispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    @property
    def urls(self) -> List[urls.path]:
        return self._urls

    @require_http_methods(['GET'])
    def _generate_spec(self, request: HttpRequest) -> HttpResponse:
        assert self._spec is not None, "spec is not set"

        endpoint_path = utils.remove_suffix(request.path, suffix=self._spec.path)
        schema = self._spec.schema(path=endpoint_path, methods_map={'': self._dispatcher.registry.values()})

        return HttpResponse(
            json.dumps(schema, cls=specs.JSONEncoder),
            content_type='application/json',
        )

    @require_http_methods(['GET'])
    def _ui_index_page(self, request: HttpRequest) -> HttpResponse:
        assert self._spec is not None and self._spec.ui is not None, "spec is not set"

        app_path = request.path.rsplit(self._spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, self._spec.path)

        return HttpResponse(self._spec.ui.get_index_page(spec_url=spec_full_path), content_type='text/html')

    @csrf_exempt
    @require_http_methods(['POST'])
    def _rpc_handle(self, request: HttpRequest) -> HttpResponse:
        """
        Handles JSON-RPC request.

        :param request: http request
        :returns: http response
        """

        if request.content_type not in pjrpc.common.REQUEST_CONTENT_TYPES:
            return HttpResponse(status=415)

        try:
            request_text = request.body.decode(request.encoding or 'utf8')
        except UnicodeDecodeError:
            return HttpResponseBadRequest()

        response = self._dispatcher.dispatch(request_text, context=request)
        if response is None:
            return HttpResponse()
        else:
            response_text, error_codes = response
            return HttpResponse(response_text, content_type=pjrpc.common.DEFAULT_CONTENT_TYPE)


class JsonRPCSites(django.utils.functional.LazyObject):
    def _setup(self) -> None:
        self._wrapped = []

        for path, endpoint in getattr(settings, 'JSONRPC_ENDPOINTS', {}).items():
            json_encoder = import_string(endpoint.get('JSON_ENCODER', 'pjrpc.server.dispatcher.JSONEncoder'))
            json_decoder = import_string(endpoint.get('JSON_DECODER', 'json.JSONDecoder'))
            spec = import_string(endpoint['SPEC']) if endpoint.get('SPEC', None) else None
            middlewares = [import_string(middleware) for middleware in endpoint.get('MIDDLEWARES', [])]
            error_handlers = {
                error: import_string(handler)
                for error, handler in endpoint.get('ERROR_HANDLERS', {}).items()
            }

            site = JsonRPCSite(
                path=path,
                spec=spec,
                json_encoder=json_encoder,
                json_decoder=json_decoder,
                middlewares=middlewares,
                error_handlers=error_handlers,
            )

            method_registry = import_string(endpoint['METHOD_REGISTRY'])
            site.dispatcher.add_methods(method_registry)

            self._wrapped.append(site)


jsonrpc_sites = JsonRPCSites()



from typing import Any, Dict, Iterable, Optional

import jsonschema

from pjrpc.common.typedefs import JsonRpcParams, MethodType
from pjrpc.server.typedefs import ExcludeFunc

from . import base


class JsonSchemaValidator(base.BaseValidator):
    """
    Parameters validator based on `jsonschema <https://python-jsonschema.readthedocs.io/en/stable/>`_ library.

    :param kwargs: default jsonschema validator arguments
    :param exclude_param: a function that decides if the parameters must be excluded
                          from validation (useful for dependency injection)
    """

    def __init__(self, exclude_param: Optional[ExcludeFunc] = None, **kwargs: Any):
        super().__init__(exclude_param=exclude_param)
        kwargs.setdefault('types', {'array': (list, tuple)})
        self.default_kwargs = kwargs

    def validate_method(
        self, method: MethodType, params: Optional['JsonRpcParams'], exclude: Iterable[str] = (), **kwargs: Any,
    ) -> Dict[str, Any]:
        """
        Validates params against method using ``pydantic`` validator.

        :param method: method to validate parameters against
        :param params: parameters to be validated
        :param exclude: parameter names to be excluded from validation
        :param kwargs: jsonschema validator arguments

        :raises: :py:class:`pjrpc.server.validators.ValidationError`
        """

        arguments = super().validate_method(method, params, exclude)

        try:
            kwargs = {**self.default_kwargs, **kwargs}
            jsonschema.validate(arguments, **kwargs)
        except jsonschema.ValidationError as e:
            raise base.ValidationError(str(e)) from e

        return arguments
